class Value(object):
    def __init__(self):
        return

    def __repr__(self):
        to_str = '{0}: '.format(type(self).__name__)
        for key, value in self.__dict__.items():
            to_str += '{0}={1}, '.format(key, value)
        return to_str

    def equals(self, object):
        if isinstance(object, Value):
            return self.value == object.value
        return False

    def getValueType(self):
        return 0

    def toObject(self):
        return self

    def write(self, dout):
        return

    def read(self, din):
        return self



import io
import struct
from whatap.pack.pack_enum import PackEnum

from whatap.value.value_enum import ValueEnum


class DataInputX(object):
    def __init__(self, v):
        self.buffer = io.BytesIO(v)

    @staticmethod
    def toInt(buf, pos):
        ch1 = buf[pos] & 0xff
        ch2 = buf[pos + 1] & 0xff
        ch3 = buf[pos + 2] & 0xff
        ch4 = buf[pos + 3] & 0xff
        return ((ch1 << 24) + (ch2 << 16) + (ch3 << 8) + (ch4 << 0))

    def readPack(self):
        type = self.readShort()
        return PackEnum.create(type).read(self)

    def readValue(self):
        type = self.readByte()
        return ValueEnum.create(type).read(self)

    # TODO
    # def available(self):
    #     return len(self.buffer.getvalue())- self.offset > 0

    # TODO
    # def readStep(self):
    #     raise
    #     # byte type = this.inner.readByte();
    #     # return StepEnum.create(type).read(this);
    #     return None

    def readIntBytes(self, max):
        ln = self.readInt()
        if ln < 0 or ln > max:
            raise Exception('error')
        return self.buffer.read(ln)

    def readBoolean(self):
        v = self.buffer.read(1)
        return struct.unpack('>?', v)[0]

    def readByte(self):
        v = self.buffer.read(1)
        return struct.unpack('>b', v)[0]

    def readShort(self):
        v = self.buffer.read(2)
        return struct.unpack('>h', v)[0]

    def readInt3(self):
        s = self.buffer.read(3)
        chrs = struct.unpack('>bBB', s)
        ch1 = (chrs[0])
        ch2 = (chrs[1])
        ch3 = (chrs[2])
        return (ch1 << 16) + (ch2 << 8) + ch3

    def readInt(self):
        v = self.buffer.read(4)

        return struct.unpack('>i', v)[0]

    def readLong5(self):
        s = self.buffer.read(5)
        chrs = struct.unpack('>bBBBB', s)
        ch1 = (chrs[0])
        ch2 = (chrs[1])
        ch3 = (chrs[2])
        ch4 = (chrs[3])
        ch5 = (chrs[4])
        return ((ch1 << 32) + (ch2 << 24) + (ch3 << 16) + (ch4 << 8) + (ch5))

    def readLong(self):
        v = self.buffer.read(8)
        return struct.unpack('>q', v)[0]

    def readFloat(self):
        v = self.buffer.read(4)
        return struct.unpack('>f', v)[0]

    def readDouble(self):
        v = self.buffer.read(8)
        return struct.unpack('>d', v)[0]

    def readDecimal(self):
        ln = self.readByte()
        if ln == 0:
            return 0
        elif ln == 1:
            return self.readByte()
        elif ln == 2:
            return self.readShort()
        elif ln == 3:
            return self.readInt3()
        elif ln == 4:
            return self.readInt()
        elif ln == 5:
            return self.readLong5()
        else:
            return self.readLong()

    def read(self, ln):
        return self.buffer.read(ln)

    def readBlob(self):
        baselen = self.readByte() & 0xff
        if baselen == 255:
            ln = self.readShort() & 0xffff
            return self.buffer.read(ln)
        elif baselen == 254:
            ln = self.readInt()
            return self.buffer.read(ln)
        elif baselen == 0:
            return []
        else:
            return self.buffer.read(baselen)

    def readText(self):
        arr = self.readBlob()
        if not len(arr):
            return ''
        else:
            return arr.decode("utf-8")

    def toByteArray(self):
        return self.buffer.getvalue()

    def readFloatArray(self):
        ln = self.readShort()
        data = []
        for _ in range(ln):
            data.append(self.readFloat())
        return data

    def readIntArray(self):
        ln = self.readShort()
        data = []
        for _ in range(ln):
            data.append(self.readInt())
        return data

    def readLongArray(self):
        ln = self.readShort()
        data = []
        for _ in range(ln):
            data.append(self.readLong())
        return data

    def readDecimalArray(self):
        ln = self.readShort()
        data = []
        for _ in range(ln):
            data.append(self.readDecimal())
        return data



import threading
import time

from whatap.conf.configure import Configure as conf
from whatap.trace.trace_context_manager import TraceContextManager
from whatap.util.date_util import DateUtil
from whatap.util.hexa32 import Hexa32

from whatap.util.linked_map import LinkedMap
from resource import getrusage, RUSAGE_SELF


class TraceContext(object):
    transfer_id = None
    transfer_info = None

    def __init__(self):
        print('TraceContext.__init__')
        self.host = ''
        self.elapsed = 0

        self.isStaticContents = 'false'

        self.id = TraceContextManager.getId()
        self.thread = threading.current_thread()
        self.thread_id = TraceContextManager.start(self)

        self.start_time = 0
        self.start_cpu = 0
        self.start_malloc = 0

        self.status = 0

        self.service_hash = 0
        self.service_name = ''
        self.remoteIp = ''
        self.error = 0
        self.error_step = ''
        self.http_method = ''
        self.http_query = ''
        self.http_content_type = ''

        self.sql_count = 0
        self.sql_time = 0
        self.sql_insert = 0
        self.sql_update = 0
        self.sql_delete = 0
        self.sql_select = 0
        self.sql_others = 0

        self.executed_sqlhash = 0
        self.active_sqlhash = 0
        self.active_dbc = 0
        self.active_crud = 0

        self.httpc_checked = False
        self.httpc_count = 0
        self.httpc_time = 0
        self.httpc_url = ''

        self.active_httpc_hash = 0
        self.httpc_host = ''
        self.httpc_port = 0

        self.mtid = 0
        self.mdepth = 0
        self.mcallee = 0

        self.mcaller_txid = 0
        self.mcaller_pcode = 0
        self.mcaller_spec = ''
        self.mcaller_url = ''

        self.userid = ''
        self._rawuserid = ''
        self.userAgent = 0
        self.userAgentString = ''
        self.referer = ''
        self.login = ''
        self.userTransaction = 0
        self.debug_sql_call = False
        self.lastSqlStep = None
        self.profileActive = 0

        self.jdbc_updated = False
        self.jdbc_update_record = 0
        self.jdbc_identity = 0
        self.jdbc_commit = 0
        self.resultSql = LinkedMap()

        self.rs_count = 0
        self.rs_time = 0
        self.db_opening = False
        self.socket_connecting = False

        self.lctx = {}

    def getElapsedTime(self, time=None):
        if not time:
            time = DateUtil.now()
        return time - self.start_time

    def getCpuTime(self):
        return int(time.time())

    def getMemory(self):
        # https://docs.python.org/3/library/resource.html?highlight=resource#resource-usage
        return int(getrusage(RUSAGE_SELF)[3] + getrusage(RUSAGE_SELF)[4])

    def resetStartTime(self):
        self.start_time = 0

    def transfer(self):
        if self.transfer_id:
            return self.transfer_id

        sb = []
        sb.append(Hexa32.toString32(self.mtid))
        sb.append(str(self.mdepth + 1))
        sb.append(Hexa32.toString32(self.id))
        transfer_id = ','.join(sb)
        return transfer_id

    def transferInfo(self):
        if self.transfer_info:
            return self.transfer_info

        sb = []
        sb.append(str(conf.PCODE))
        sb.append(str(conf.mtrace_spec))
        if conf.mtrace_send_url_length:
            sb.append(self.service_name[:conf.mtrace_send_url_length])

        transfer_info = ','.join(sb)
        return transfer_info

    def setTransfer(self, headerString):
        x = headerString.find(',')
        if x > 0:
            self.mtid = Hexa32.toLong32(headerString[0:x])
            y = headerString.find(',', x + 1)
            if y > 0:
                self.mdepth = int(headerString[x + 1:y])
                z = headerString.find(',', y + 1)

                if z < 0:
                    self.mcaller_txid = Hexa32.toLong32(headerString[y + 1:])
                else:
                    self.mcaller_txid = Hexa32.toLong32(headerString[y + 1: z])

    def setTransferInfo(self, headerString):
        x = headerString.index(',')
        s1 = headerString[0: x]
        self.mcaller_pcode = 0 if len(s1) == 0 else s1

        y = headerString.index(',', x + 1)
        self.mcaller_spec = headerString[x + 1: y]
        self.mcaller_url = headerString[y + 1:]

    def setTxid(self, myid):
        self.id = Hexa32.toLong32(myid)




from whatap.conf.configure import Configure as conf
from whatap.net.packet_type_enum import PacketTypeEnum
from whatap.net.udp_session import UdpSession
from whatap.trace.mod.application_wsgi import interceptor, trace_handler, \
    interceptor_error
from whatap.trace.trace_context_manager import TraceContextManager
from whatap.trace.trace_context import TraceContext
from whatap.util.date_util import DateUtil
from whatap.util.userid_util import UseridUtil as userid_util
import whatap.util.throttle_util as throttle_util
import time

def blocking_handler():
    def handler(func):
        from django.http import HttpResponse
        def wrapper(*args, **kwargs):
            if conf.throttle_enabled and args and len(args) > 2:
                remote_ip = userid_util.getRemoteAddr(args)
                req = args[1]
                path = req['PATH_INFO']
                if throttle_util.isblocking(remote_ip, path):
                    if conf.reject_event_enabled:
                        ctx = TraceContextManager.getLocalContext()
                        if not ctx:
                            ctx = TraceContext()

                        throttle_util.sendrejectevent(ctx, path, remote_ip)
                    if conf.throttle_blocked_forward:
                        response = HttpResponse(status=302)
                        response['Location'] = throttle_blocked_forward
                        return response
                    status = '403 Forbidden'
                    start_response= args[2]
                    start_response(status, [])
                    response = HttpResponse(content=conf.throttle_blocked_message, status=403)
                    return response
            return func(*args, **kwargs)

        return wrapper

    return handler

def instrument(module):
    def wrapper(fn):
        @trace_handler(fn, True)
        @blocking_handler()
        def trace(*args, **kwargs):

            callback = interceptor(fn, *args, **kwargs)

            return callback
        
        return trace
    
    module.WSGIHandler.__call__ = wrapper(module.WSGIHandler.__call__)



def instrument_handlers_base(module):
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            request = args[1]
            ctx = TraceContextManager.getLocalContext()
            callback = fn(*args, **kwargs)
            
            if conf.trace_auto_normalize_enabled:
                resolver_match = request.resolver_match
                if resolver_match:
                    
                    path = request.path
                    for key, value in resolver_match.kwargs.items():
                        path = path.replace(resolver_match.kwargs[key],
                                            '{' + key + '}')
                    
                    start_time = DateUtil.nowSystem()
                    ctx.start_time = start_time
                    ctx.service_name = path
                    
                    if hasattr(resolver_match, 'view_name'):
                        type_name = 'View' \
                            if resolver_match._func_path != resolver_match.view_name \
                            else 'Function'
                        desc = '{0}: {1}'.format(type_name,
                                                 resolver_match._func_path)
                        datas = [' ', ' ', desc]
                        ctx.elapsed = DateUtil.nowSystem() - start_time
                        UdpSession.send_packet(PacketTypeEnum.TX_MSG, ctx,
                                               datas)

            return callback
        
        return trace

    if hasattr(module.BaseHandler, 'apply_response_fixes'):
        module.BaseHandler.apply_response_fixes = wrapper(
            module.BaseHandler.apply_response_fixes)

    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            callback = fn(*args, **kwargs)
            
            e = args[3]
            status_code = callback.status_code
            errors = [e[0].__name__,
                      e[1].args[1] if len(
                          e[1].args) > 1 \
                          else repr(e[1].args[0])]
            interceptor_error(status_code, errors)
            
            return callback
        
        return trace

    if hasattr(module.BaseHandler, 'handle_uncaught_exception'):
        module.BaseHandler.handle_uncaught_exception = wrapper(
            module.BaseHandler.handle_uncaught_exception)

    def get_response_wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            callback = fn(*args, **kwargs)
            request = args[1]
            ctx = TraceContextManager.getLocalContext()
            if ctx and conf.trace_user_enabled:
                if not conf.trace_user_using_ip:
                    userid_util.setUserId(request, callback, ctx._rawuserid )

            return callback

        return trace
    module.BaseHandler.get_response = get_response_wrapper(
        module.BaseHandler.get_response)


def instrument_generic_base(module):
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            self = args[0]
            ctx = TraceContextManager.getLocalContext()
            start_time = DateUtil.nowSystem()
            ctx.start_time = start_time
            desc = '{0}.{1}'.format(self.__module__, type(self).__name__)
            datas = [' ', ' ', desc]
            
            ctx.elapsed = DateUtil.nowSystem() - start_time
            UdpSession.send_packet(PacketTypeEnum.TX_MSG, ctx, datas)
            
            callback = fn(*args, **kwargs)
            return callback
        
        return trace
    
    module.View.dispatch = wrapper(module.View.dispatch)




# Django==1.10

def instrument_urls_base(module):
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            callback = fn(*args, **kwargs)
            return callback
        
        return trace
    
    module.reverse = wrapper(module.reverse)


def instrument_handlers_exception(module):
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            callback = fn(*args, **kwargs)
            return callback
        
        return trace
    
    module.convert_exception_to_response.convert_exception_to_response = wrapper(
        module.convert_exception_to_response.convert_exception_to_response)

def instrument_handlers_static(module):
    def get_response_wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            callback = fn(*args, **kwargs)

            ctx = TraceContextManager.getLocalContext()
            if ctx:
                ctx.userid = 0

            return callback

        return trace

    module.StaticFilesHandler.get_response = get_response_wrapper(
        module.StaticFilesHandler.get_response)




class ParamDef(object):
	KEEPALIVE = 1
	AGENT_BOOT_ENV = 2
	CONFIGURE_UPDATE = 3
	CONFIGURE_GET = 4
	COMPO_VERSIONS = 5
	THREAD_LIST = 7
	THREAD_DETAIL = 8
	GET_ACTIVE_STACK = 9
	HEAP_HISTO = 10
	LOADED_CLASS_LIST = 11
	GET_ENV = 12
	SYSTEM_GC = 13
	SET_CONFIG = 14
	OPEN_SOCKET_LIST = 15
	LOADED_CLASS_DETAIL = 16
	LOADED_CLASS_REDEFINE = 17
	AGENT_JAR_LIST = 18
	AGENT_JAR_SAVE = 19
	AGENT_JAR_DELETE = 20
	RESET_STRING_SENT_MARK = 21
	AGENT_LOG_LIST = 22
	AGENT_LOG_READ = 23
	THREAD_CONTROL = 24
	GET_ACTIVE_TRANSACTION_LIST = 25
	GET_ACTIVE_TRANSACTION_DETAIL = 26
	GET_TOPOLOGY = 27
	INFRA_NET_STAT=28
	INFRA_PS_EF=29
	INFRA_DOCKER_LIST=30
	INFRA_REMOTE_CMD=33
	INFRA_PEER_LIST = 34
	INFRA_AGENT_UPDATE = 36
	AGENT_DUMP_LIST = 31
	AGENT_DUMP_READ = 32
	AGENT_STAT = 35
	JVM_THREAD_DUMP = 37
	JVM_HEAP_DUMP = 38
	KUBENETES = 39
	GET_ACTIVE_STATS = 40
	MODULE_DEPENDENCY = 101
	DBX_INFO = 201
	DBX_SETTINGS = 202
	DBX_USER_SCRIPT = 207
	DBX_SCRIPT_LIST = 208
	DBX_SCRIPT_SQL = 209
	DBX_KILL_SESSION = 210
	DBX_GET_ALERT = 211
	DBX_SET_ALERT = 212
	DBX_PLAN = 213
	DBX_SESSION_LIST = 214
	DBX_SET_CPUCORES = 215
	DBX_SESSION_STAT = 203
	DBX_SESSION_EVENT = 204
	DBX_TOPSTAT = 205
	DBX_TOPEVENT = 206
	CONFIG_DBC = 301
	METHOD_PERF_STAT = 302
	BLOCKING_DETECT = 303
	CW_INFO = 401
	CW_EVENTS = 402
	CW_CREDENTIAL = 403
	CW_REGIONS = 404
	CW_METRIC_NAMES = 405
	CW_AWS_RESOURCE_IDS = 406
	CW_STATISTICS = 407
	CW_UNITS = 408
	CW_METRICS = 409


import io
import math
import struct

from whatap import PY2, PY3
from whatap.pack.pack import Pack

from whatap.value.null_value import NullValue

BYTE_MIN_VALUE = -128
BYTE_MAX_VALUE = 127
SHORT_MIN_VALUE = -32768
SHORT_MAX_VALUE = 32767
INT3_MIN_VALUE = -0x800000
INT3_MAX_VALUE = 0x007fffff
INT_MIN_VALUE = -0x80000000
INT_MAX_VALUE = 0x7fffffff
LONG5_MIN_VALUE = -0x8000000000
LONG5_MAX_VALUE = 0x0000007fffffffff
LONG_MIN_VALUE = -0x8000000000000000
LONG_MAX_VALUE = 0x7fffffffffffffff


class DataOutputX(object):
    def __init__(self, size=None):
        if size:
            self.buffer = io.BytesIO(bytearray(size))
        else:
            self.buffer = io.BytesIO()

    @staticmethod
    def toInt(buf, pos):
        ch1 = buf[pos] & 0xff
        ch2 = buf[pos + 1] & 0xff
        ch3 = buf[pos + 2] & 0xff
        ch4 = buf[pos + 3] & 0xff
        return ((ch1 << 24) + (ch2 << 16) + (ch3 << 8) + (ch4 << 0))

    def writePack(self, v, ln_fmt):
        self.writeShort(v.getPackType())
        v.write(self)
        if ln_fmt:
            remainder = len(self.buffer.getvalue()) % ln_fmt
            if remainder:
                self.write(bytearray(int(math.floor(ln_fmt - remainder))))
        return self

    def writeStep(self, step):
        self.writeByte(step.getStepType())
        step.write(self)
        return self

    @classmethod
    def toBytes(cls, v, ln_fmt=None):
        if isinstance(v, Pack):
            return cls().writePack(v, ln_fmt).toByteArray()

        buf = bytearray(4)
        buf[0] = ((v % 0x100000000) >> 24) & 0xff
        buf[1] = ((v % 0x100000000) >> 16) & 0xff
        buf[2] = ((v % 0x100000000) >> 8) & 0xff
        buf[3] = ((v % 0x100000000) >> 0) & 0xff
        return buf

    @classmethod
    def toBytesLong(cls, v):
        buf = bytearray(8)
        buf[0] = ((v % 0x100000000) >> 56) & 0xff
        buf[1] = ((v % 0x100000000) >> 48) & 0xff
        buf[2] = ((v % 0x100000000) >> 40) & 0xff
        buf[3] = ((v % 0x100000000) >> 32) & 0xff
        buf[4] = ((v % 0x100000000) >> 24) & 0xff
        buf[5] = ((v % 0x100000000) >> 16) & 0xff
        buf[6] = ((v % 0x100000000) >> 8) & 0xff
        buf[7] = ((v % 0x100000000) >> 0) & 0xff
        return buf

    def set(self, dest, pos, src):
        dest[pos:pos + len(src)] = src[0:]
        return dest

    def writeIntBytes(self, b):
        if not b or not len(b):
            self.writeInt(0)
        else:
            self.writeInt(len(b))
            self.write(b)
        return self

    def writeBoolean(self, v):
        self.buffer.write(struct.pack('>?', v))
        return self

    def writeByte(self, v):
        v = v & 0xFF
        self.buffer.write(struct.pack('>B', v))
        return self

    def writeShort(self, v):
        v = v & 0xFFFF
        self.buffer.write(struct.pack('>H', v))
        return self

    def writeInt3(self, v):
        v1 = (v >> 16) & 0xFF
        v2 = (v >> 8) & 0xFF
        v3 = (v >> 0) & 0xFF
        self.buffer.write(struct.pack('>BBB', v1, v2, v3))
        return self

    def writeInt(self, v):
        v = v & 0xFFFFFFFF
        self.buffer.write(struct.pack('>I', v))
        return self

    def writeLong5(self, v):
        v1 = ((v >> 32) & 0xFF)
        v2 = ((v >> 24) & 0xFF)
        v3 = ((v >> 16) & 0xFF)
        v4 = ((v >> 8) & 0xFF)
        v5 = ((v >> 0) & 0xFF)
        self.buffer.write(struct.pack('>BBBBB', v1, v2, v3, v4, v5))
        return self

    def writeLong(self, v):
        v = v & 0xFFFFFFFFFFFFFFFF
        self.buffer.write(struct.pack('>Q', v))
        return self

    def writeFloat(self, v):
        self.buffer.write(struct.pack('>f', v))
        return self

    def writeDouble(self, v):
        self.buffer.write(struct.pack('>d', v))
        return self

    def writeDecimal(self, v):
        if v == 0:
            self.writeByte(0)
        elif BYTE_MIN_VALUE <= v <= BYTE_MAX_VALUE:
            self.writeByte(1)
            self.writeByte(v)
        elif SHORT_MIN_VALUE <= v <= SHORT_MAX_VALUE:
            self.writeByte(2)
            self.writeShort(v)
        elif INT3_MIN_VALUE <= v <= INT3_MAX_VALUE:
            self.writeByte(3)
            self.writeInt3(v)
        elif INT_MIN_VALUE <= v <= INT_MAX_VALUE:
            self.writeByte(4)
            self.writeInt(v)
        elif LONG5_MIN_VALUE <= v <= LONG5_MAX_VALUE:
            self.writeByte(5)
            self.writeLong5(v)
        elif LONG_MIN_VALUE <= v <= LONG_MAX_VALUE:
            self.writeByte(8)
            self.writeLong(v)
        return self

    def write(self, v):
        self.buffer.write(v)
        return self

    def writeBlob(self, v):
        ln = len(v)
        if not v or not ln:
            self.writeByte(0)
        else:
            if ln <= 253:
                self.writeByte(ln)
                self.write(v)
            elif ln <= 65535:
                self.writeByte(255)
                self.writeShort(ln)
                self.write(v)
            else:
                self.writeByte(254)
                self.writeInt(ln)
                self.write(v)
        return self

    def writeText(self, v):
        if not v:
            self.writeByte(0)
        else:
            self.writeBlob(v.encode("utf-8"))
        return self

    def writeUTF(self, v):
        v = v.encode('utf-8')
        if len(v) > 65535:
            v = v[:65535]

        self.buffer.write(struct.pack('>H', len(v)))
        self.buffer.write(v)
        return self

    def writeValue(self, v):
        if not v:
            v = NullValue()

        self.writeByte(v.getValueType())
        v.write(self)
        return self

    def toByteArray(self):
        return self.buffer.getvalue()

    def flush(self):
        self.buffer.flush()

    def writeFloatArray(self, vv):
        if not vv:
            self.writeShort(0)
        else:
            self.writeShort(len(vv))
            for v in vv:
                self.writeFloat(v)
        return self

    def writeIntArray(self, vv):
        if not vv:
            self.writeShort(0)
        else:
            self.writeShort(len(vv))
            for v in vv:
                self.writeInt(v)
        return self

    def writeLongArray(self, vv):
        if not vv:
            self.writeShort(0)
        else:
            self.writeShort(len(vv))
            for v in vv:
                self.writeLong(v)
        return self

    def writeDecimalArray(self, vv):
        if not vv:
            self.writeShort(0)
        else:
            self.writeShort(len(vv))
            for v in vv:
                self.writeDecimal(v)
        return self

    def writeToPos(self, pos, v):
        if PY3:
            struct.pack_into('>I', self.buffer.getbuffer(), pos, v)
        else:
            b = io.BytesIO()
            b.write(struct.pack('>I', v & 0xFFFFFFFF))

            buffer_arr = bytearray(self.buffer.getvalue())
            buffer_arr[pos:pos + 4] = b.getvalue()
            self.buffer = io.BytesIO(buffer_arr)



from whatap.trace import get_dict
from whatap.trace.mod.application_wsgi import trace_handler, \
    interceptor_step_error
from whatap.trace.trace_context_manager import TraceContextManager
from whatap.net.udp_session import UdpSession
from whatap.net.packet_type_enum import PacketTypeEnum
from whatap.util.date_util import DateUtil
import traceback

def intercept_connect(fn, *args, **kwargs):
    ctx = TraceContextManager.getLocalContext()
    start_time = DateUtil.nowSystem()
    ctx.start_time = start_time

    try:
        callback = fn(*args, **kwargs)
        return callback
    except Exception as e:
        interceptor_step_error(e)
        raise
    finally:
        text = None

        if kwargs:
            text = 'smtp://'
            text += kwargs.get('host')
            text += ':'
            text += str(kwargs.get('port', 0))
        elif args and len(args) > 2:
            text = 'smtplilb.SMTP.connect('
            text += args[1]
            text += ','
            text += str(args[2])
            text += ')'
        if text:
            payloads = [text, '']
            ctx.elapsed = DateUtil.nowSystem() - start_time
            UdpSession.send_packet(PacketTypeEnum.TX_METHOD, ctx, payloads)

def intercept_method(method, fn, *args, **kwargs):
    ctx = TraceContextManager.getLocalContext()

    start_time = DateUtil.nowSystem()
    ctx.start_time = start_time

    try:
        callback = fn(*args, **kwargs)
        return callback
    except Exception as e:
        interceptor_step_error(e)
        raise
    finally:
        payloads = [method, '']
        ctx.elapsed = DateUtil.nowSystem() - start_time
        UdpSession.send_packet(PacketTypeEnum.TX_METHOD, ctx, payloads)

def instrument_smtp(module):
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            callback = intercept_connect(fn, *args, **kwargs)
            return callback

        return trace

    if hasattr(module, 'SMTP') and hasattr(module.SMTP, 'connect'):
        module.SMTP.connect= wrapper(module.SMTP.connect)

    def wrapper(fn, method):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            callback = intercept_method(method, fn, *args, **kwargs)
            return callback

        return trace

    for method in ['sendmail',]:
        if hasattr(module, 'SMTP') and hasattr(module.SMTP, method):
            setattr(module.SMTP, method, wrapper(getattr(module.SMTP, method), 'smtplib.SMTP.{}'.format(method)))



import os, sys
import random
import threading
import time
import uuid
import random
from whatap import logging
from whatap.conf.configure import Configure as conf
from whatap.io.data_inputx import DataInputX
from whatap.io.data_outputx import DataOutputX
from whatap.util.linked_map import LinkedMap
import whatap.util.date_util as date_util


class TraceContextManager(object):
    entry = LinkedMap()
    local = threading.local()
    node = uuid.getnode()
    clock_seq = random.randrange(1 << 14)
    
    @classmethod
    def keys(cls):
        return cls.entry.keys()
    
    @classmethod
    def getActiveCount(cls):
        act = [0 for _ in range(3)]
        try:
            en = cls.entry.values()
            while en.hasMoreElements():
                ctx = en.nextElement()
                elapsed = ctx.getElapsedTime()
                if elapsed < conf.trace_active_transaction_yellow_time:
                    act[0] += 1
                elif elapsed < conf.trace_active_transaction_red_time:
                    act[1] += 1
                else:
                    act[2] += 1
        
        except Exception as e:
            logging.debug(e, extra={'id': 'WA310'}, exc_info=True)
        
        return act

    @classmethod
    def getActiveStats(cls):
        act = [0 for _ in range(5)]
        try:
            en = cls.entry.values()
            while en.hasMoreElements():
                ctx = en.nextElement()
                if ctx.active_sqlhash:
                    act[1] += 1 # sql
                elif ctx.active_httpc_hash:
                    act[2] += 1 # httpc
                elif ctx.db_opening:
                    act[3] += 1 # dbc
                elif ctx.socket_connecting:
                    act[4] += 1  # socket
                else:
                    act[0] += 1 # method
        except Exception as e:
            logging.debug(e, extra={'id': 'WA311'}, exc_info=True)
        return act

    @classmethod
    def getContextEnumeration(cls):
        return cls.entry.values()
    
    @classmethod
    def getContext(cls, key):
        return cls.entry.get(key)
    
    @classmethod
    def getLocalContext(cls):
        # print('TraceContextManager.getLocalContext:', cls.getCurrentKey())
        return cls.getContext(cls.getCurrentKey())
        
        return cls.local.context
    
    @classmethod
    def getId(cls):
        uuid1 = uuid.uuid1(TraceContextManager.node, TraceContextManager.clock_seq)
        key = (uuid1.int >> 64 & 0xffffffffffffffff) ^ (uuid1.int << 64 & 0xffffffffffffffff)
        return key

    @classmethod
    def getCurrentKey(cls):
        thread_id = cls.getCurrentThreadId()
        pid = os.getpid()
        dout = DataOutputX()
        dout.writeInt(pid)
        dout.writeInt(0)
        din = DataInputX(dout.buffer.getvalue())
        key = din.readLong() + thread_id

        return key

    @classmethod
    def start(cls, o):
        key = cls.getCurrentKey()
        print('TraceContextManager.start:', key)
        cls.local.context = o
        cls.entry.put(key, o)

        return key
    
    @classmethod
    def parseThreadId(cls, key):
        pid = os.getpid()
        dout = DataOutputX()
        dout.writeInt(pid)
        dout.writeInt(0)
        din = DataInputX(dout.buffer.getvalue())
        thread_id = key - din.readLong()
        return thread_id, pid
    
    @classmethod
    def end(cls, key):
        cls.local.context = None
        cls.entry.remove(str(key))

    @classmethod
    def getTxProfile(cls, n):
        ctx = cls.getLocalContext()
        if not ctx:
            return None
        return ctx.profile.getLastSteps(n)

    @classmethod
    def getCurrentThreadId(cls):
        greenlet = sys.modules.get('greenlet')

        if greenlet:
            current = greenlet.getcurrent()
            if current is not None and current.parent:
                return id(current.parent)
        return threading.get_ident()


PLUGIN = {}
IMPORT_HOOKS = {}



DEFINITION = {
    'plugin': [
        ('', 'instrument_plugin'),
    ],
    'httpc_httplib': [
        ('httplib', 'instrument_httplib'),
        ('http.client', 'instrument_httplib'),
        ('httplib2', 'instrument_httplib2'),
    ],
    'httpc_requests': [
        ('requests.sessions', 'instrument_requests'),
    ],
    'httpc_urllib3': [
        ('urllib3.request', 'instrument_urllib3'),
        # ('requests.packages.urllib3.request', 'instrument_urllib3'),
    ],
    'httpc_django': [
        ('revproxy.views', 'instrument_revproxy_views'),
    ],
    'database_mysql': [
        ('MySQLdb', 'instrument_MySQLdb'),
        ('MySQLdb.cursors', 'instrument_MySQLdb_cursors'),
        ('pymysql', 'instrument_pymysql'),
        ('pymysql.cursors', 'instrument_pymysql_cursors'),
    ],
    'database_postgresql': [
        ('psycopg2', 'instrument_psycopg2'),
        ('psycopg2._psycopg', 'instrument_psycopg2_connection'),
        ('psycopg2.extensions', 'instrument_psycopg2_extensions'),
    ],
    'database_toolkit': [
        ('sqlalchemy.orm.session', 'instrument_sqlalchemy'),
        # ('sqlalchemy.engine.default', 'instrument_sqlalchemy_engine'),
    ],
    'application_wsgi': [
        ('', ''),
    ],
    'application_bottle': [
        ('bottle', 'instrument'),
    ],
    'application_cherrypy': [
        ('cherrypy', 'instrument'),
    ],
    'application_django': [
        ('django.core.handlers.wsgi', 'instrument'),
        ('django.core.handlers.base', 'instrument_handlers_base'),
        ('django.views.generic.base', 'instrument_generic_base'),
        ('django.contrib.staticfiles.handlers', 'instrument_handlers_static'),

        # Django==1.10
        ('django.urls.resolvers', 'instrument_url_resolvers', False),
        ('django.urls.base', 'instrument_urls_base', False),
        ('django.core.handlers.exception', 'instrument_handlers_exception',
         False),
    ],
    'application_flask': [
        ('flask', 'instrument'),
    ],
    'application_tornado': [
        ('tornado.web', 'instrument'),
    ],
    'application_celery': [
        # ('celery.task.base', 'instrument_task_base'),
        #('celery.app.task', 'instrument_app_task'),
        #('celery.worker', 'instrument_celery_worker'),
        #('celery.concurrency.processes', 'instrument_celery_worker'),
        #('celery.concurrency.prefork', 'instrument_celery_worker'),
        ('celery.execute.trace', 'instrument_celery_execute_trace'),
        ('celery.task.trace', 'instrument_celery_execute_trace'),
        ('celery.app.trace', 'instrument_celery_execute_trace'),
        #('billiard.pool', 'instrument_billiard_pool'),
    ],
    'database_redis':[
        ('redis','instrument_redis_connection'),
    ],
    'email_smtp':[
        ('smtplib','instrument_smtp'),
    ],
    # 'amqp_kombu':[
    #     ('kombu','instrument_kombu'),
    # ],
    'amqp_pika':[
        ('pika.channel','instrument_pika'),
    ],
    'application_nameko': [
        ('nameko.containers', 'instrument_nameko_spawn_worker'),
        ('spsengine.containers', 'instrument_nameko_spawn_worker'),
    ],

}



from whatap.trace import get_dict
from whatap.trace.mod.application_wsgi import trace_handler, \
    interceptor_step_error, start_interceptor, end_interceptor
from whatap.trace.trace_context import TraceContext
from whatap.trace.trace_context_manager import TraceContextManager
from whatap.net.udp_session import UdpSession
from whatap.net.packet_type_enum import PacketTypeEnum
from whatap.util.date_util import DateUtil

def parseConnection(conn):

    connstr = ''
    if conn.host:
        connstr += conn.host
    if conn.virtual_host:
        connstr += "/" + conn.virtual_host
    return connstr

def intercept_publish(fn, *args, **kwargs):
    ctx = TraceContextManager.getLocalContext()

    start_time = DateUtil.nowSystem()
    ctx.start_time = start_time
    text = 'rabbitmq://'
    text += parseConnection(args[0].connection.params)
    text += '?exchage='
    text += kwargs['exchange']
    text += '&routing='
    text += kwargs['routing_key']
    ctx.active_dbc = text
    ctx.lctx['dbc'] = text

    ctx.active_dbc = 0
    ctx.db_opening = True
    ctx.elapsed = DateUtil.nowSystem() - start_time
    datas = [' ', ' ', 'MQ SESSION INFO: ' + text]
    UdpSession.send_packet(PacketTypeEnum.TX_MSG, ctx, datas)


start_time = DateUtil.nowSystem()
    ctx.start_time = start_time

    try:
        callback = fn( *args, **kwargs)
        return callback
    except Exception as e:
        interceptor_step_error(e)
    finally:
        ctx.db_opening = False

    return callback

def instrument_pika(module):
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            callback = intercept_publish(fn, *args, **kwargs)
            return callback
        
        return trace

    if hasattr(module, 'Channel') and hasattr(module.Channel, 'basic_publish'):
        module.Channel.basic_publish = wrapper(module.Channel.basic_publish)





from functools import wraps
from whatap.trace.trace_context import TraceContext
from whatap.trace.trace_context_manager import TraceContextManager
from whatap.trace.mod.application_wsgi import start_interceptor, end_interceptor, interceptor_step_error
from whatap import logging
from whatap.conf.configure import Configure as conf
import time

def interceptor(fn, task_name, *args, **kwargs):
    now = time.time()
    if now - conf.last_loaded > conf.init_load_interval/1000:
        conf.init(False)
    ctx = TraceContext()
    ctx.service_name = task_name
    start_interceptor(ctx)
    try:
        callback = fn(*args, **kwargs)
        ctx = TraceContextManager.getLocalContext()
        return callback
    except Exception as e:
        interceptor_step_error(e)
        raise e
    finally:
        if ctx:
            end_interceptor(ctx=ctx)

def trace_handler(fn, start=False):
    def handler(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            ctx = TraceContextManager.getLocalContext()
            if not start and not ctx:
                return fn(*args, **kwargs)
            if ctx and ctx.error_step:
                end_interceptor(ctx=ctx)
                raise Exception(ctx.error_step)
            callback = func(*args, **kwargs)
            return callback
        return wrapper

    return handler

def instrument_celery_execute_trace(module):
    def wrapper(fn, task_name):
        @trace_handler(fn, start=True)
        def trace(*args, **kwargs):
            callback = interceptor(fn, task_name, *args, **kwargs)
            return callback

        return trace
    if hasattr(module, 'build_tracer'):
        _build_tracer = module.build_tracer
        def build_tracer(name, task,*args, **kwargs):
            task = task or module.tasks[name]
            name = ''
            if hasattr(task,'name'):
                name = task.name
            task.run = wrapper(task.run, name)
            return _build_tracer(name, task,*args, **kwargs)

        module.build_tracer = build_tracer


import threading

import sys, os
import traceback
import time
import logging
import socket
import errno

from whatap.io import DataOutputX
from whatap.net.packet_enum import PacketEnum
from whatap.net.packet_type_enum import PacketTypeEnum
from whatap.net.param_def import ParamDef
from whatap.trace import TraceContextManager


class UdpSession(object):
    s = None
    buffer_arr = []
    thread_lock = threading.Lock()
    read_timeout = 5

    @classmethod
    def udp(cls):
        import socket, struct
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, PacketEnum.PACKET_SEND_BUFFER_SIZE)
        sock.connect((PacketEnum.SERVER, PacketEnum.PORT))
        cls.s = sock
        sock.settimeout(cls.read_timeout)

        return sock

    @classmethod
    def send_packet(cls, packet_type, ctx, datas=[]):
        import whatap.net.udp_thread
        try:
            if len(b''.join(cls.buffer_arr)) > PacketEnum.PACKET_BUFFER_SIZE:
                cls.send(packet_type, ctx)

            if not datas:
                return

            dout = DataOutputX()
            start_header_buffer_size = len(dout.buffer.getvalue())

            # header
            dout.writeByte(packet_type)
            dout.writeInt(PacketEnum.PACKET_VERSION)
            dout.writeInt(0)

            # body
            start_body_buffer_size = len(dout.buffer.getvalue())

            if ctx:
                dout.writeLong(ctx.id)
                dout.writeLong(ctx.start_time)
                dout.writeInt(ctx.elapsed)
                dout.writeLong(ctx.getCpuTime())
                dout.writeLong(ctx.getMemory())
                dout.writeInt(os.getpid())
                dout.writeLong(ctx.thread_id)

            diff = PacketEnum.PACKET_BUFFER_SIZE \
                   - len(b''.join(cls.buffer_arr)) - (PacketEnum.PACKET_BODY_REQUIRED_SIZE + 1)
            for data in datas:
                data = str(data)
                diff -= len(data)
                if diff < 0:
                    data = ' '
                    logging.debug('message too long.', extra={'id': 'WA999'},
                                  exc_info=False)
                dout.writeUTF(data)

            with cls.thread_lock:
                dout.writeToPos(
                    start_header_buffer_size + PacketEnum.PACKET_HEADER_LEN_POS,
                    len(dout.buffer.getvalue()) - start_body_buffer_size)

                cls.buffer_arr.append(dout.buffer.getvalue())

                if not ctx or packet_type == PacketTypeEnum.TX_START \
                        or packet_type == PacketTypeEnum.TX_END:
                    cls.send(packet_type, ctx)
        except Exception as e:
            logging.debug(e, extra={'id': 'WA913'}, exc_info=True)

    @classmethod
    def send(cls, packet_type, ctx):
        s = cls.s
        if not s:
            s = cls.udp()

        # for name, value in PacketTypeEnum.__dict__.items():
        #     if value == packet_type:
        #         print('pack_type: ', extra={'id': 'WA919'}, exc_info=True)

        try:
            s.sendall(b''.join(cls.buffer_arr))
        except ConnectionRefusedError as e:
            logging.debug(e, extra={'id': 'WA911'}, exc_info=True)
            cls.udp()
        except Exception as e:
            logging.debug(len(b''.join(cls.buffer_arr)))
            cls.buffer_arr.pop()
            logging.debug(len(b''.join(cls.buffer_arr)))
            logging.debug(e, extra={'id': 'WA912'}, exc_info=True)
            cls.send(packet_type, ctx)
            logging.debug('re send!')
        finally:
            cls.buffer_arr = []
            if packet_type == PacketTypeEnum.TX_END:
                TraceContextManager.end(ctx.thread_id)

    @classmethod
    def get(cls):
        from whatap.conf.configure import Configure as conf
        try:
            received = cls.s.recv(1024)
            if received:
                cls.handle(received.decode().split(','))
        except socket.error as serr:
            if serr.errno == errno.EAGAIN:
                conf.init(display=False)
            else:
                raise serr

    @classmethod
    def handle(cls, received):
        param_id, request, extra = tuple(received)
        param_id = int(param_id)
        thread_id = 0
        pid = 0
        if(extra != ''):
            extra_datas = extra.replace(' ', ', ').split(', ')
            thread_id, pid = TraceContextManager.parseThreadId(int(extra_datas[0]))

        data = ''
        if not param_id:
            # active stack pack
            datas = []
            data = extra.replace(' ', ', ')

            frame = sys._current_frames().get(thread_id)
            if not frame:
                return

            for stack in traceback.extract_stack(frame):
                line = stack[0]
                line_num = stack[1]
                method_name = stack[2]

                if line.find('/whatap/trace') > -1 or line.find('/threading.py') > -1:
                    continue
                data += '{} ({}:{})\n'.format(method_name, line, line_num)

            datas.append(data)

            cls.send_packet(PacketTypeEnum.ACTIVE_STACK, None, datas)
            stats = TraceContextManager.getActiveStats()
            datas = [','.join([str(x) for x in stats])]
            cls.send_packet(PacketTypeEnum.ACTIVE_STATS, None, datas)
        else:
            # param pack
            # format: "[packetType], [ctx], [datas: xxxx xxxx xxxx]"
            datas = [str(param_id), request]
            if param_id == ParamDef.MODULE_DEPENDENCY:
                import pip
                data = ', '.join(map(str, list(pip.get_installed_distributions())))
                datas.append(data)
            elif param_id == ParamDef.GET_ACTIVE_TRANSACTION_DETAIL:
                data = extra.replace(' ', ', ')

                frame = sys._current_frames().get(thread_id)
                if not frame:
                    return

                for stack in traceback.extract_stack(frame):
                    line = stack[0]
                    line_num = stack[1]
                    method_name = stack[2]

                    if line.find('whatap/trace') > -1:
                        continue
                    data += '{} ({}:{})\n'.format(method_name, line, line_num)

                datas.append(data)
            elif param_id == ParamDef.GET_ACTIVE_STATS:
                stats = TraceContextManager.getActiveStats()
                datas = [','.join([str(x) for x in stats])]
                cls.send_packet(PacketTypeEnum.ACTIVE_STATS, None, datas)
                return

            cls.send_packet(PacketTypeEnum.TX_PARAM, None, datas)



from whatap.trace.mod.application_wsgi import transfer, trace_handler, \
    interceptor_httpc_request, interceptor_sock_connect


def instrument_httplib(module):
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            # set mtid header
            kwargs['headers'] = transfer(kwargs.get('headers', {}))
            
            # set httpc_url
            httpc_url = args[1]
            callback = interceptor_httpc_request(fn, httpc_url, *args, **kwargs)
            return callback
        
        return trace
    
    module.HTTPConnection.request = wrapper(module.HTTPConnection.request)
    module.HTTPConnection.connect = wrapper(interceptor_sock_connect)

def instrument_httplib2(module):
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            # set mtid header
            kwargs['headers'] = transfer(kwargs.get('headers', {}))
            
            # set httpc_url
            httpc_url = args[1]
            callback = interceptor_httpc_request(fn, httpc_url, *args, **kwargs)
            return callback
        
        return trace
    
    module.Http.request = wrapper(module.Http.request)
    module.HTTPConnectionWithTimeout.connect = wrapper(interceptor_sock_connect)
    module.HTTPSConnectionWithTimeout.connect = wrapper(interceptor_sock_connect)


import os, time

from whatap.conf.configuration import Configuration


class Configure(object):
    PCODE = 0
    dev = False
    net_udp_port = 6600
    last_loaded = 0
    last_config_modified = 0
    observers = []
    POD_NAME = os.environ.get("POD_NAME") if os.environ.get('POD_NAME') else os.environ.get("PODNAME") if os.environ.get("PODNAME") else ''

    @classmethod
    def init(cls, display=True):

        cls.last_loaded = time.time()
        return cls.load(display)

    @classmethod
    def load(cls, display=True):
        home = 'WHATAP_HOME'
        try:
            whatap_config = os.path.join(os.environ[home],os.environ['WHATAP_CONFIG'])
            last_modified = os.path.getmtime(whatap_config)
            if cls.last_config_modified >= last_modified:
                return True
            for key, value in Configuration.items():
                setattr(cls, key, value)
            with open(whatap_config, 'r') as f:
                for line in f:
                    cls.last_config_modified = last_modified
                    line_strip = line.strip()
                    if not line_strip or line_strip.startswith('#'):
                        continue
                    try:
                        key, value = line.split('=')
                        key = key.strip()
                        value = value.strip()
                        cls.setProperty(key, value)
                    
                    except Exception as e:
                        print('WHATAP: ', e)
                        continue
            for callback in cls.observers:
                callback()
        except Exception as e:
            from whatap import CONFIG_FILE_NAME, init_config
            init_config(home)
            return False
        else:
            return True
        finally:
            if display:
                from whatap import Logger
                Logger()
            
    @classmethod
    def getProperty(cls, key, value=None):
        if hasattr(cls, key):
            return getattr(cls, key)
        else:
            return value
    
    @classmethod
    def setProperty(cls, name, value):
        if hasattr(cls, name):
            if isinstance(getattr(cls, name), bool) and str(value) != 'true':
                value = False
        
        setattr(cls, name, value)
    
    def getStringSet(cls, key, default_value, deli):
        l = list()
        value = cls.getProperty(key, default_value)
        if value:
            for v in value.split(deli):
                l.append(v)
        return l

    @classmethod
    def addObserver(cls, callback):
        cls.observers.append(callback)


class PacketEnum(object):
    PACKET_BUFFER_SIZE = 48 * 1024
    PACKET_SEND_BUFFER_SIZE = 64 * 1024

    SERVER = "127.0.0.1"
    PORT = 6600
    PACKET_VERSION = 20101

    # PACKET_HEADER
    PACKET_HEADER_TYPE_POS = 0
    PACKET_HEADER_TYPE_SIZE = 1

    PACKET_HEADER_VERSION_POS = 1
    PACKET_HEADER_VERSION_SIZE = 4

    PACKET_HEADER_LEN_POS = 5
    PACKET_HEADER_LEN_SIZE = 4

    # PACKET_BODY
    PACKET_BODY_ID_POS = 9
    PACKET_BODY_ID_SIZE = 8

    PACKET_BODY_START_TIME_POS = 17
    PACKET_BODY_START_TIME_SIZE = 8

    PACKET_BODY_ELAPSED_POS = 25
    PACKET_BODY_ELAPSED_SIZE = 4

    PACKET_BODY_CPU_POS = 29
    PACKET_BODY_CPU_SIZE = 8

    PACKET_BODY_MEMOERY_POS = 37
    PACKET_BODY_MEMOERY_SIZE = 8

    PACKET_BODY_THREAD_ID_POS = 45
    PACKET_BODY_THREAD_ID_SIZE = 8

    PACKET_BODY_TRACE_DATA_POS = 53

    # PACKET_SIZE
    PACKET_HEADER_SIZE = PACKET_HEADER_TYPE_SIZE + PACKET_HEADER_VERSION_SIZE + PACKET_HEADER_LEN_SIZE
    PACKET_BODY_REQUIRED_SIZE = PACKET_BODY_ID_SIZE + PACKET_BODY_START_TIME_SIZE + PACKET_BODY_ELAPSED_SIZE \
                + PACKET_BODY_CPU_SIZE + PACKET_BODY_MEMOERY_SIZE + PACKET_BODY_THREAD_ID_SIZE



import threading
import whatap.io.data_outputx as data_outputx
from io import StringIO

class IntSetry(object):
	def __init__(self, key=0, next= None):
		self.key = key
		self.next = next

	def clone(self):
		return  IntSetry(self.key, None if self.next is None else next.clone())

	def getKey(self):
		return self.key

	def equals(self, o):
		if not instance(o,IntSetry):
			return False

		return self.key == o.getKey()

	def hashCode(self):
		return self.key

	def toString(self):
		return str(key)

class Enumer(object):
	def __init__(self, table):
		self.table = table
		self.index = len(table)
		self.entry = None
		self.lastReturned = None

	def hasMoreElements(self):
		while self.entry is None and self.index > 0:
			self.index -= 1
			self.entry = self.table[self.index]

		return self.entry is not None

	def nextInt(self):
		while self.entry is None and self.index > 0:
			self.index -= 1
			self.entry = self.table[self.index]
		if self.entry is not None:
			e = self.lastReturned = self.entry
			self.entry = e.next
			return e.key
		raise Exception("no more next")


class EmptyEnumer(object):
	def hasMoreElements(self):
		return 0

	def nextInt(self):
		return False

class IntSet(object):
	emptyEnumer = EmptyEnumer()

	def __init__(self, initCapacity = int(101), loadFactor = float(0.75)):
		self.count = 0
		if initCapacity < 0:
			raise Exception("Capacity Error: " + initCapacity)
		if loadFactor <= 0:
			raise Exception("Load Count Error: " + loadFactor)
		if initCapacity == 0:
			initCapacity = 1
		self.loadFactor = loadFactor;
		self.table = [None]*initCapacity
		self.threshold = int(initCapacity * loadFactor)

		self.lock = threading.Lock()

	def size(self):
		return self.count

	def values(self):
		try:
			self.lock.acquire()
			return Enumer(self.table)
		finally:
			self.lock.release()

	def contains(self, key):
		buk = self.table
		index = (key & data_outputx.INT_MAX_VALUE) % len(buk)
		e = buk[index]
		while e is not None:
			if e.key == key:
				return True
			e = e.next

		return False

	def get(self, key):
		buk = self.table
		index = (key & data_outputx.INT_MAX_VALUE) % len(buk)

		e = buk[index]
		while e is not None:
			if e.key == key:
				return e.key

		return 0

	def rehash(self):
		oldCapacity = len(self.table)
		oldMap = self.table
		newCapacity = oldCapacity * 2 + 1
		newMap = [None]*newCapacity
		self.threshold = int(newCapacity * loadFactor)
		self.table = newMap

		for i in range(oldCapacity, 1, -1):
			old = oldMap[i]
			while old is not None:
				e = old
				old = old.next
				index = (e.key & data_outputx.INT_MAX_VALUE) % newCapacity
				e.next = newMap[index]
				newMap[index] = e

	def putAll(self, values):
		if not values :
			return
		for v in values:
			self.put(v)

	def put(self, value):
		buk = self.table
		index = (value & data_outputx.INT_MAX_VALUE) % len(buk)
		e = buk[index]
		while  e is not None:
			if e.key == value:
				return False
			e = e.next
		if self.count >= self.threshold:
			self.rehash()
			buk = self.table
			index = (value & Integer.MAX_VALUE) % len(buk)

		e = IntSetry(value, buk[index])
		buk[index] = e
		self.count +=1
		return True

	def remove(self, key):
		buk = self.table
		index = (key & data_outputx.INT_MAX_VALUE) % len(buk)
		e = buk[index]
		prev = None
		while e is not None:
			if e.key == key:
				if prev is not None:
					prev.next = e.next
				else:
					buk[index] = e.next
				self.count-=1
				return key
			prev = e, e = e.next

		return 0

	def clear(self):
		buk = table
		for index in range(len(buk), 0, -1):
			buk[index] = None
		self.count = 0

	def toString(self):
		max = self.size() - 1
		buf = StringIO()
		it = this.values()
		buf.write("{")
		for i in range(max):
			key = it.nextInt()
			buf.write(str(key))
			if i < max:
				buf.write(", ")
		buf.write("}")

		return buf.getvalue()

	def toArray(self):
		_keys = int[self.size()]
		en = self.values()
		for i in range(len(_keys)):
			_keys[i] = en.nextInt()
		return _keys

	def isExistElseAdd(self, value):
		return self.put(value) == False

	def putStrHash(self, s):
		if s is not None:
			self.put(hash(s))



from whatap.trace.mod.application_wsgi import trace_handler, \
    interceptor, interceptor_error


def instrument(module):
    def wrapper(fn):
        @trace_handler(fn, True)
        def trace(*args, **kwargs):
            request = args[0].request
            
            environ = dict()
            environ['HTTP_HOST'] = request.host
            environ['PATH_INFO'] = request.path
            environ['QUERY_STRING'] = request.query_arguments
            environ['REMOTE_ADDR'] = request.remote_ip
            environ['HTTP_USER_AGENT'] = request.headers.get('User-Agent', '')
            environ['HTTP_REFERER'] = ''
            
            callback = interceptor((fn, environ), *args, **kwargs)
            
            return callback
        
        return trace
    
    module.RequestHandler._execute = wrapper(module.RequestHandler._execute)
    
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            callback = fn(*args, **kwargs)
            
            e = args[1]
            status_code = args[0]._status_code
            errors = [e.__class__.__name__, str(e)]
            interceptor_error(status_code, errors)
            return callback
        
        return trace
    
    module.RequestHandler._handle_request_exception = wrapper(
        module.RequestHandler._handle_request_exception)



from datetime import datetime


class Pack(object):
    def __init__(self):
        self.pcode = 0
        self.oid = 0
        self.time = 0
    
    def __repr__(self):
        to_str = '{0}: '.format(type(self).__name__)
        
        data = self.__dict__.copy()
        for key, value in data.items():
            if key == 'time':
                value = datetime.fromtimestamp(value / 1000) \
                    .strftime('%Y-%m-%d %H:%M:%S')
                
            to_str += '{0}={1}, '.format(key, value)
        return to_str
    
    def getPackType(self):
        return 0
    
    def write(self, dout):
        dout.writeDecimal(self.pcode)
        dout.writeInt(self.oid)
        dout.writeLong(self.time)
    
    def read(self, din):
        self.pcode = din.readDecimal()
        self.oid = din.readInt()
        self.time = din.readLong()
        return self



import traceback

import sys

from whatap.conf.configure import Configure as conf
from whatap.net.packet_type_enum import PacketTypeEnum
from whatap.net.udp_session import UdpSession
from whatap.trace.trace_context import TraceContext
from whatap.trace.trace_context_manager import TraceContextManager
from whatap.util.date_util import DateUtil
from whatap.util.userid_util import UseridUtil as userid_util
from functools import wraps
from whatap import logging
import re

from whatap.util.hexa32 import Hexa32

import traceback, threading

def isEventlet():
    greenlet = sys.modules.get('greenlet')

    if greenlet:
        current = greenlet.getcurrent()
        if current is not None and current.parent:
            print('thread:',threading.get_ident(),'greenlet:', id(current.parent))
            traceback.print_stack()
            return True
    else:
        return False

def trace_handler(fn, start=False):
    def handler(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # if isEventlet():
            #     print("trace_handler: greenlet found")
            # else:
            #     print("trace_handler: greenlet not found")
            ctx = TraceContextManager.getLocalContext()
            # print("trace_handler ctx:", ctx)
            if not start and not ctx:
                return fn(*args, **kwargs)
            
            # check raise step error
            if ctx and ctx.error_step:
                end_interceptor(ctx=ctx)
                raise Exception(ctx.error_step)
            
            try:
                callback = func(*args, **kwargs)
            except Exception as e:
                logging.debug(e, extra={'id': 'WA917'}, exc_info=True)
                print(e, dict(extra={'id': 'WA917'}, exc_info=True))
                return fn(*args, **kwargs)
            else:
                return callback
        
        return wrapper
    
    return handler


def start_interceptor(ctx):
    if conf.dev:
        logging.debug('start transaction id(seq): {}'.format(ctx.id),
                      extra={'id': 'WA111'})
        print('start transaction id(seq): {}'.format(ctx.id),dict(extra={'id': 'WA111'}))
    
    start_time = DateUtil.nowSystem()
    ctx.start_time = start_time
    
    datas = [ctx.host,
             ctx.service_name,
             ctx.remoteIp,
             ctx.userAgentString,
             ctx.referer,
             ctx.userid,
             ctx.isStaticContents
             ]
    # print("start_interceptor txid:", ctx.id)
    UdpSession.send_packet(PacketTypeEnum.TX_START, ctx, datas)
    
    return ctx

def end_interceptor(thread_id=None, ctx= None):
    if not ctx:
        ctx = TraceContextManager.getContext(
            thread_id) if thread_id else TraceContextManager.getLocalContext()
    if not ctx:
        return

    if conf.dev:
        logging.debug('end   transaction id(seq): {}'.format(ctx.id),
                      extra={'id': 'WA112'})
        print('end   transaction id(seq): {}'.format(ctx.id),
                      dict(extra={'id': 'WA112'}))

    start_time = DateUtil.nowSystem()
    ctx.start_time = start_time

    datas = [ctx.host, ctx.service_name, ctx.mtid, ctx.mdepth, ctx.mcaller_txid,
             ctx.mcaller_pcode, ctx.mcaller_spec, ctx.mcaller_url]
    ctx.elapsed = DateUtil.nowSystem() - start_time

    UdpSession.send_packet(PacketTypeEnum.TX_END, ctx, datas)

def interceptor(rn_environ, *args, **kwargs):
    if not isinstance(rn_environ, tuple):
        rn_environ = (rn_environ, args[1])
    fn, environ = rn_environ
    
    ctx = TraceContext()
    ctx.host = environ.get('HTTP_HOST', '').split(':')[0]
    ctx.service_name = environ.get('PATH_INFO', '')

    ctx.remoteIp = userid_util.getRemoteAddr(args)

    ctx.userAgentString = environ.get('HTTP_USER_AGENT', '')
    ctx.referer = environ.get('HTTP_REFERER''', '')

    if conf.trace_user_enabled:
        if conf.trace_user_using_ip:
            ctx.userid = userid_util.getRemoteAddr(args)
        else:
            ctx.userid, ctx._rawuserid = userid_util.getUserId(args, ctx.remoteIp)

    mstt = environ.get('HTTP_{}'.format(
        conf._trace_mtrace_caller_key.upper().replace('-', '_')), '')
    
    if mstt:
        ctx.setTransfer(mstt)
        if conf.stat_mtrace_enabled:
            val = environ.get('HTTP_{}'.format(
                conf._trace_mtrace_info_key.upper().replace('-', '_')), '')
            if val and len(val):
                ctx.setTransferInfo(val)
            pass

        myid = environ.get('HTTP_{}'.format(
            conf._trace_mtrace_callee_key.upper().replace('-', '_')), '')
        if myid:
            ctx.setTxid(myid)
    start_interceptor(ctx)
    
    callback = fn(*args, **kwargs)

    if ctx:
        query_string = environ.get('QUERY_STRING', '')
        if query_string:
            ctx.service_name += '?{}'.format(query_string)

        if ctx.service_name.find('.') > -1 and ctx.service_name.split('.')[
            1] in conf.web_static_content_extensions:
            ctx.isStaticContents = 'true'

        if getattr(callback, 'status_code', None):
            status_code = callback.status_code
            errors = [callback.reason_phrase, callback.__class__.__name__]
            interceptor_error(status_code, errors)
        
        if conf.profile_http_header_enabled:
            keys = []
            for key, value in environ.items():
                if key.startswith('HTTP_'):
                    keys.append(key)
            keys.sort()
            
            text = ''
            for key in keys:
                text += '{}={}\n'.format(key.split('HTTP_')[1].lower(),
                                         environ[key])
            
            datas = ['HTTP-HEADERS', 'HTTP-HEADERS', text]
            ctx.start_time = DateUtil.nowSystem()
            UdpSession.send_packet(PacketTypeEnum.TX_MSG, ctx, datas)
        end_interceptor(ctx=ctx)
    return callback


def interceptor_error(status_code, errors):
    print('error', status_code, errors)
    ctx = TraceContextManager.getLocalContext()
    ctx.status = int(status_code / 100)
    if ctx.status >= 4 and not ctx.error:
        ctx.error = 1
        
        error = ''
        frame = sys._current_frames().get(ctx.thread.ident)
        if not frame:
            return
        
        for stack in traceback.extract_stack(frame):
            line = stack[0]
            line_num = stack[1]
            method_name = stack[2]

            if line.find('/whatap/trace') > -1 or line.find(
                    '/threading.py') > -1:
                continue
            error += '{} ({}:{})\n'.format(method_name, line, line_num)
        
        errors.append(error)
        
        # errors.append(''.join(traceback.format_list(traceback.extract_stack(sys._current_frames()[ctx.thread.ident]))))
        UdpSession.send_packet(PacketTypeEnum.TX_ERROR, ctx, errors)


def interceptor_step_error(e):
    print('interceptor_step_error', e)
    ctx = TraceContextManager.getLocalContext()
    ctx.error_step = e
    if not ctx.error:
        ctx.error = 1
    
    errors = []
    errors.append(e.__class__.__name__)
    errors.append(str(e.args[0]))
    
    error = ''
    frame = sys._current_frames().get(ctx.thread.ident)
    if not frame:
        return
    
    for stack in traceback.extract_stack(frame):
        line = stack[0]
        line_num = stack[1]
        method_name = stack[2]
        
        if line.find('/whatap/trace') > -1 or line.find('/threading.py') > -1:
            continue
        error += '{} ({}:{})\n'.format(method_name,line, line_num)
    
    errors.append(error)
    # errors.append(''.join(traceback.format_list(traceback.extract_stack(sys._current_frames()[ctx.thread.ident]))))
    UdpSession.send_packet(PacketTypeEnum.TX_ERROR, ctx, errors)

    if conf.profile_exception_stack:
        desc = '\n'.join(errors)
        datas = [' ', ' ', desc]
        ctx.start_time = DateUtil.nowSystem()
        UdpSession.send_packet(PacketTypeEnum.TX_MSG, ctx, datas)

def interceptor_httpc_request(fn, httpc_url, *args, **kwargs):
    ctx = TraceContextManager.getLocalContext()
    if not ctx or ctx.active_httpc_hash:
        return fn(*args, **kwargs)

    param = None
    method = None
    if httpc_url.find('?') > -1:
        httpc_url, param = httpc_url.split('?')

    start_time = DateUtil.nowSystem()
    ctx.start_time = start_time
    ctx.httpc_url = httpc_url
    ctx.active_httpc_hash = ctx.httpc_url
    
    try:
        callback = fn(*args, **kwargs)
        return callback
    except Exception as e:
        interceptor_step_error(e)
        print('interceptor_httpc_request', e)
    finally:
        datas = [ctx.httpc_url, ctx.mcallee]
        ctx.elapsed = DateUtil.nowSystem() - start_time
        UdpSession.send_packet(PacketTypeEnum.TX_HTTPC, ctx, datas)

        if conf.profile_http_parameter_enabled:
            if type(args[1]) == dict:
                param = (args[1].body if 'body' in args[1] else args[1]) or param
                method = args[1].method if 'method' in args[1] else args[1]

            if param:
                datas = ['HTTP-PARAMETERS', method, param]
                ctx.start_time = DateUtil.nowSystem()
                UdpSession.send_packet(PacketTypeEnum.TX_SECURE_MSG, ctx, datas)

        ctx.active_httpc_hash = 0
        ctx.httpc_url = None

def interceptor_sock_connect(fn, *args, **kwargs):
    ctx = TraceContextManager.getLocalContext()
    if not ctx :
        return fn(*args, **kwargs)

    try:
        ctx.socket_connecting = True
        callback = fn(*args, **kwargs)
        return callback
    except Exception as e:
        interceptor_step_error(e)
        print('interceptor_sock_connect', e)
    finally:
        ctx.socket_connecting = False

def interceptor_db_con(fn, db_type, *args, **kwargs):
    ctx = TraceContextManager.getLocalContext()
    if ctx.db_opening:
        return fn(*args, **kwargs)
    
    start_time = DateUtil.nowSystem()
    ctx.start_time = start_time

    ctx.db_opening = True
    try:
        callback = fn(*args, **kwargs)
    finally:

        ctx.db_opening = False

    if not kwargs:
        kwargs = dict(
            x.split('=') for x in re.sub(r'\s*=\s*', '=', args[0]).split())
    
    text = '{}://'.format(db_type)
    text += kwargs.get('user')
    text += "@"
    text += kwargs.get('host')
    text += '/'
    text += kwargs.get('database', kwargs.get('db', kwargs.get('dbname')))
    ctx.active_dbc = text
    ctx.lctx['dbc'] = text
    
    ctx.active_dbc = 0

    datas = [text]
    ctx.elapsed = DateUtil.nowSystem() - start_time
    UdpSession.send_packet(PacketTypeEnum.TX_DB_CONN, ctx, datas)
    
    return callback

def addQuote(arg_dict):
    quoted_dict = dict()

    for k, v in arg_dict.items():
        if isinstance(v, str):
            quoted_dict[k] = "'" + v.replace("'", "\\'") + "'"
        else:
            quoted_dict[k] = v

    return quoted_dict

def interceptor_db_execute(fn, *args, **kwargs):
    # print('interceptor_db_execute step -1')
    ctx = TraceContextManager.getLocalContext()
    # print('interceptor_db_execute step -2 ctx:', ctx)
    self = args[0]
    query = None
    if len(args) > 2 and type(args[2]) == dict and args[2]:
        try:
            query = args[1] % addQuote(args[2])
        except Exception as e:
            pass
    try:
        if not query:
            query = args[1].decode()
    except Exception as e:
        query = args[1]
    #print('interceptor_db_execute step -3', query)
    if not query or ctx.active_sqlhash:
        #print('interceptor_db_execute step -3.1', ctx.active_sqlhash)
        return fn(*args, **kwargs)


    # print('interceptor_db_execute step -4')
    start_time = DateUtil.nowSystem()
    ctx.start_time = start_time
    ctx.active_sqlhash = query
    # print('interceptor_db_execute step -5')
    try:
        callback = fn(*args, **kwargs)
        # print('interceptor_db_execute step -6')
        return callback
    except Exception as e:

        interceptor_step_error(e)
    finally:
        # print('interceptor_db_execute step -7 txid:', ctx.id)
        datas = [ctx.lctx.get('dbc', ''), query]
        ctx.elapsed = DateUtil.nowSystem() - start_time
        UdpSession.send_packet(PacketTypeEnum.TX_SQL, ctx,
                               datas)
        # print('interceptor_db_execute step -8')
        count = self.rowcount
        # print('interceptor_db_execute step -9', count)
        if count > -1:
            desc = '{0}: {1}'.format('Fetch count', count)
            datas = [' ', ' ', desc]
            ctx.elapsed = 0
            UdpSession.send_packet(PacketTypeEnum.TX_MSG, ctx, datas)
        
        ctx.active_sqlhash = 0
        # print('interceptor_db_execute step -10')

def interceptor_db_close(fn, *args, **kwargs):
    ctx = TraceContextManager.getLocalContext()
    ctx.db_opening = False
    
    if not conf.profile_dbc_close:
        return fn(*args, **kwargs)
    
    start_time = DateUtil.nowSystem()
    ctx.start_time = start_time
    
    callback = fn(*args, **kwargs)
    
    text = 'DB: Close Connection.'
    datas = [' ', ' ', text]
    ctx.elapsed = DateUtil.nowSystem() - start_time
    UdpSession.send_packet(PacketTypeEnum.TX_MSG, ctx, datas)
    return callback


check_seq = 1


def inter_tx_trace_auto_on(ctx):
    try:
        if isinstance(conf.mtrace_rate, str):
            conf.mtrace_rate = int(conf.mtrace_rate)
    except ValueError:
        conf.mtrace_rate = 0
    finally:
        if conf.mtrace_rate <= 0 or ctx.httpc_checked or ctx.mtid != 0:
            return
        
        ctx.httpc_checked = True
        
        try:
            inter_tx_trace_auto_on.check_seq += 1
        except AttributeError:
            inter_tx_trace_auto_on.check_seq = 1
        finally:
            check_seq = inter_tx_trace_auto_on.check_seq
            
            rate = int(conf.mtrace_rate / 10)
            if rate == 10:
                ctx.mtid = TraceContextManager.getId()
            elif rate == 9:
                if check_seq % 10 != 0:
                    ctx.mtid = TraceContextManager.getId()
            elif rate == 8:
                if check_seq % 5 != 0:
                    ctx.mtid = TraceContextManager.getId()
            elif rate == 7:
                if check_seq % 4 != 0:
                    ctx.mtid = TraceContextManager.getId()
            elif rate == 6:
                if check_seq % 3 != 0:
                    ctx.mtid = TraceContextManager.getId()
            elif rate == 5:
                if check_seq % 2 == 0:
                    ctx.mtid = TraceContextManager.getId()
            elif rate == 4:
                if check_seq % 3 == 0 or check_seq % 5 == 0:
                    ctx.mtid = TraceContextManager.getId()
            elif rate == 3:
                if check_seq % 4 == 0 or check_seq % 5 == 0:
                    ctx.mtid = TraceContextManager.getId()
            elif rate == 2:
                if check_seq % 5 == 0:
                    ctx.mtid = TraceContextManager.getId()
            elif rate == 1:
                if check_seq % 10 == 0:
                    ctx.mtid = TraceContextManager.getId()


def transfer(headers):
    ctx = TraceContextManager.getLocalContext()
    
    if not ctx.mtid:
        inter_tx_trace_auto_on(ctx)
    
    if ctx.mtid:
        headers[conf._trace_mtrace_caller_key] = ctx.transfer()
        if conf.stat_mtrace_enabled:
            headers[conf._trace_mtrace_info_key] = ctx.transferInfo()

        ctx.mcallee = TraceContextManager.getId()
        headers[conf._trace_mtrace_callee_key] = Hexa32.toString32(ctx.mcallee)
    return headers



from whatap.value.value import Value
from whatap.value.value_enum import ValueEnum


class BooleanValue(Value):
    def __init__(self, value=None):
        super(BooleanValue, self).__init__()
        self.value = value if isinstance(value, bool) else False

    def compareTo(self, object):
        if isinstance(object, BooleanValue):
            if self.value == object.value:
                return 0
            return 1 if self.value else -1
        return 1

    def equals(self, object):
        if isinstance(object, BooleanValue):
            return self.value == object.value
        return False

    def hashCode(self):
        return 1 if self.value else 0

    def write(self, dout):
        dout.writeBoolean(self.value)

    def read(self, din):
        self.value = din.readBoolean()
        return self

    def getValueType(self):
        return ValueEnum.BOOLEAN



from whatap.conf.configure import Configure as conf
from whatap.pack.pack_enum import EventLevel
from whatap.net.udp_session import UdpSession
from whatap.util.hash_util import HashUtil
from whatap.util.int_set import IntSet
import whatap.util.string_util as string_util
from whatap.util.ip_util import IPUtil
from whatap.net.packet_type_enum import PacketTypeEnum

def getThrottleStringToHashSet(value, deli):
    intset = IntSet()
    if value is not None:
        vv = string_util.tokenizer(value, deli)
        if vv is None:
            return intset
        for x in vv:
            intset.put(HashUtil.hash(x))

    return intset

def getThrottleIP(value, deli):
    intset = IntSet()
    if value is not None:
        vv = string_util.tokenizer(value, deli)
        if not vv :
            return intset
        for x in vv:
            ip = IPUtil.toInt(x)
            if ip != 0:
                intset.put(ip)

    return intset

def getThrottleIgnorePrefixl(value, deli):
    strset = set()
    if value:
        vv = StringUtil.tokenizer(value, deli)
        if not vv:
            return []
        for x in vv:
            if x:
                strset.add(x)

    return list(strset)


passingUrlSet = getThrottleStringToHashSet(conf.throttle_passing_url, ",")
passingPrefix = getThrottleIgnorePrefixl(conf.throttle_passing_url_prefix, ",")
blockingUrlSet = getThrottleStringToHashSet(conf.throttle_blocking_url, ",")
blockingIPSet = getThrottleIP(conf.throttle_blocking_ip, ",")

blocking_enabled = blockingUrlSet.size() > 0 or blockingIPSet.size() > 0

def getHash():
    k1 = string_util.trimEmpty(conf.throttle_passing_url)
    k2 = string_util.trimEmpty(conf.throttle_passing_url_prefix)
    k3 = string_util.trimEmpty(conf.throttle_blocking_url)
    k4 = string_util.trimEmpty(conf.throttle_blocking_ip)
    return hash(k1) ^ hash(k2) ^ hash(k3) ^ hash(k4)


def isblocking(str_ip, path):
    remote_ip = IPUtil.toInt(str_ip)
    if not blocking_enabled:
        return False
    if blockingIPSet.contains(remote_ip):
        return True
    urlhash = HashUtil.hashFromString(path)
    if blockingUrlSet.contains(urlhash):
        return True
    return False

def sendrejectevent(ctx, path, ip):
    print('sendrejectevent ', path, ip)
    pathHash = HashUtil.hashFromString(path)
    datas = ("REJECTED_URL", EventLevel.WARNING,
             "Rejected " + path,
             pathHash,
             path,
             ip,
    )
    print('sendrejectevent datas', datas)
    UdpSession.send_packet(PacketTypeEnum.EVENT, ctx, datas)
    print('sendrejectevent ', path, ip,' complete')

valueHash = 0
def updateConfig():
    global passingUrlSet, passingPrefix, blockingUrlSet, blockingIPSet, blocking_enabled, ignoreContext, ignoreDomain, valueHash
    newHash = getHash()
    if valueHash != newHash:
        passingUrlSet = getThrottleStringToHashSet(conf.throttle_passing_url, ",")
        passingPrefix = getThrottleIgnorePrefixl(conf.throttle_passing_url_prefix, ",")
        blockingUrlSet = getThrottleStringToHashSet(conf.throttle_blocking_url, ",")
        blockingIPSet = getThrottleIP(conf.throttle_blocking_ip, ",")

        blocking_enabled = len(blockingUrlSet) > 0 or len(blockingIPSet) > 0

        ignoreContext = StringUtil.tokenizer(conf.throttle_blocking_ignore_context, ",")
        ignoreDomain = getThrottleStringToHashSet(conf.throttle_blocking_ignore_domain, ",")

        valueHash = getHash()
conf.addObserver(updateConfig)






def tokenizer(target, delim):
    if not target:
        return None

    tokens = []
    for t in target.split(delim):
        t = t.strip()
        if t:
            tokens.append(t)

    return tokens

def trimEmpty(src):
    if src:
        return src.strip()
    else:
        return ''


import io

from whatap.io.data_inputx import DataInputX
from whatap.io.data_outputx import DataOutputX

from whatap.util.hexa32 import Hexa32


class Key(object):
    def __init__(self):
        self.pcode = None
        self.security_key = io.BytesIO()

    def __repr__(self):
        to_str = '{0}: '.format(type(self).__name__)

        res = []
        for key, value in self.__dict__.items():
            if type(value).__name__.find('bytes') > -1:
                dout = DataOutputX()
                dout.writeIntBytes(value)

                if isinstance(value, bytes) or isinstance(value, bytearray):
                    for o in value:
                        if o > 128:
                            o -= 256
                        res.append(o)

            to_str += '{0}={1}, '.format(key, res)
        return to_str


class License(object):
    @staticmethod
    def getKey(lic):
        tokens = lic.split('-')
        dout = DataOutputX(len(tokens) * 8)
        for token in tokens:
            dout.writeLong(Hexa32.toLong32(token))
        key = Key()
        din = DataInputX(dout.toByteArray())
        key.pcode = din.readDecimal()
        key.security_key = din.readBlob()
        return key

    @staticmethod
    def getProjectCode(lic):
        k = License.getKey(lic)
        return k.pcode



from whatap.trace.mod.application_wsgi import transfer, trace_handler, \
    interceptor_httpc_request


def instrument_revproxy_views(module):
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            callback = fn(*args, **kwargs)
            
            # set mtid header
            callback = transfer(callback)
            return callback
        return trace
    
    module.ProxyView.get_proxy_request_headers = wrapper(module.ProxyView.get_proxy_request_headers)

    
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            # set httpc_url
            httpc_url = args[0].upstream
            callback = interceptor_httpc_request(fn, httpc_url, *args, **kwargs)
            return callback

        return trace
    
    module.ProxyView.dispatch = wrapper(module.ProxyView.dispatch)




from whatap.trace import get_dict
from whatap.trace.mod.application_wsgi import trace_handler, \
    interceptor_db_con, interceptor_db_execute, interceptor_db_close


def instrument_MySQLdb(module):
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            db_type = 'mysql'
            callback = interceptor_db_con(fn, db_type, *args, **kwargs)
            return callback
        
        return trace
    
    module.connect = wrapper(module.connect)
    
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            callback = interceptor_db_close(fn, *args, **kwargs)
            return callback
        
        return trace
    
    get_dict(module.connection)['close'] = wrapper(
        module.connection.close)


def instrument_MySQLdb_cursors(module):
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            callback = interceptor_db_execute(fn, *args, **kwargs)
            return callback
        
        return trace
    
    module.BaseCursor.execute = wrapper(module.BaseCursor.execute)
    if hasattr(module.BaseCursor,'callproc'):
        module.BaseCursor.callproc= wrapper(module.BaseCursor.callproc)

def instrument_pymysql(module):
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            db_type = 'mysql'
            callback = interceptor_db_con(fn, db_type, *args, **kwargs)
            return callback
        
        return trace
    
    module.connect = wrapper(module.connect)
    
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            callback = interceptor_db_close(fn, *args, **kwargs)
            return callback
        
        return trace
    
    module.connections.Connection.close = wrapper(
        module.connections.Connection.close)

def instrument_pymysql_cursors(module):
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            callback = interceptor_db_execute(fn, *args, **kwargs)
            return callback

        return trace

    module.Cursor.execute = wrapper(module.Cursor.execute)


from whatap.trace.mod.application_wsgi import interceptor, trace_handler, \
    interceptor_error


def instrument(module):
    def wrapper(fn):
        @trace_handler(fn, True)
        def trace(*args, **kwargs):
            callback = interceptor(fn, *args, **kwargs)
            return callback
        
        return trace
    
    module.Flask.wsgi_app = wrapper(module.Flask.wsgi_app)
    
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            callback = fn(*args, **kwargs)
            if not callback:
                e = args[1]
                errors = [e.__class__.__name__]
                if hasattr(args[1], 'code'):
                    status_code = e.code
                    errors.append(e.description)
                else:
                    status_code = 500
                    errors.append(e.args[0])
                
                interceptor_error(status_code, errors)
            
            return callback
        
        return trace
    
    module.Flask._find_error_handler = wrapper(
        module.Flask._find_error_handler)



from whatap.trace import get_dict
from whatap.trace.mod.application_wsgi import trace_handler, \
    interceptor_step_error
from whatap.trace.trace_context_manager import TraceContextManager
from whatap.net.udp_session import UdpSession
from whatap.net.packet_type_enum import PacketTypeEnum
from whatap.util.date_util import DateUtil

def intercept_send_command(fn, instance, *args, **kwargs):
    ctx = TraceContextManager.getLocalContext()
    if ctx.db_opening :
        return fn(instance, *args, **kwargs)
    start_time = DateUtil.nowSystem()
    ctx.start_time = start_time

    port_or_path = getattr(instance, 'port', getattr(instance, 'path', None))
    text = 'redis://'
    text += getattr(instance, 'host','localhost')
    text += ':{}'.format(port_or_path)
    text += '/'
    text += str(getattr(instance, 'db', 0))
    ctx.active_dbc = text
    ctx.lctx['dbc'] = text

    ctx.active_dbc = 0
    ctx.db_opening = True
    datas = [text]
    ctx.elapsed = DateUtil.nowSystem() - start_time
    UdpSession.send_packet(PacketTypeEnum.TX_DB_CONN, ctx, datas)
    query = " ".join([ str(arg)[:20] for arg in args])
    if ctx.active_sqlhash:
        return fn(*args, **kwargs)

    start_time = DateUtil.nowSystem()
    ctx.start_time = start_time
    ctx.active_sqlhash = query

    try:
        callback = fn(instance, *args, **kwargs)
        return callback
    except Exception as e:
        interceptor_step_error(e)
    finally:
        ctx.db_opening = False
        datas = [ctx.lctx.get('dbc', ''), query]
        ctx.elapsed = DateUtil.nowSystem() - start_time
        UdpSession.send_packet(PacketTypeEnum.TX_SQL, ctx,
                               datas)
    return callback


def interceptor_read_response(fn, *args, **kwargs):
    ctx = TraceContextManager.getLocalContext()
    if not ctx:
        return fn(*args, **kwargs)
    start_time = DateUtil.nowSystem()
    ctx.start_time = start_time

    callback = None
    try:
        callback = fn(*args, **kwargs)
        return callback
    except Exception as e:
        interceptor_step_error(e)
    finally:
        if callback:
            count = len(callback)

            if count > -1:
                msg = '{0}: {1}'.format('Fetch bytes', count)
                datas = [ msg,' ', ' ']
                ctx.elapsed = 0
                UdpSession.send_packet(PacketTypeEnum.TX_MSG, ctx, datas)

        ctx.active_sqlhash = 0

def instrument_redis_connection(module):
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            callback = intercept_send_command(fn, args[0], *args[1:], **kwargs)
            return callback
        
        return trace

    if hasattr(module, 'Connection') and hasattr(module.Connection, 'send_command'):
        module.Connection.send_command = wrapper(module.Connection.send_command)

    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            callback = interceptor_read_response(fn, *args, **kwargs)
            return callback

        return trace

    if hasattr(module, 'Connection') and hasattr(module.Connection, 'read_response'):
        module.Connection.read_response= wrapper(module.Connection.read_response)




import ctypes

from whatap.trace.trace_context_manager import TraceContextManager

_get_dict = ctypes.pythonapi._PyObject_GetDictPtr
_get_dict.restype = ctypes.POINTER(ctypes.py_object)
_get_dict.argtypes = [ctypes.py_object]


def get_dict(obj):
    return _get_dict(obj).contents.value




class PackEnum(object):
    PARAMETER = 0x0100
    COUNTER = 0x0200
    COUNTER_1 = 0x0201
    PROFILE = 0x0300
    ACTIVESTACK = 0x0400
    TEXT = 0x0700
    ERROR_SNAP = 0x0800
    ERROR_SNAP_1 = 0x0801
    REALTIME_USER = 0x0f00
    REALTIME_USER_1 = 0x0f01

    STAT_SERVICE = 0x0900
    STAT_GENERAL = 0x0910
    STAT_SQL = 0x0a00
    STAT_HTTPC = 0x0b00
    STAT_ERROR = 0x0c00
    STAT_METHOD = 0x0e00
    STAT_TOP_SERVICE = 0x1000
    STAT_REMOTE_IP = 0x1100
    STAT_USER_AGENT = 0x1200
    STAT_USER_AGENT_1 = 0x1201

    EVENT = 0x1400
    HITMAP = 0x1500
    HITMAP_1 = 0x1501
    EXTENSION = 0x1600

    COMPOSITE = 0x1700

    @staticmethod
    def create(p):
        return None

class EventLevel(object):
    FATAL = 30
    WARNING = 20
    INFO = 10
    NONE = 0



from whatap.trace import get_dict
from whatap.trace.mod.application_wsgi import trace_handler, \
    interceptor_step_error, start_interceptor, end_interceptor
from whatap.trace.trace_context import TraceContext
from whatap.trace.trace_context_manager import TraceContextManager
from whatap.net.udp_session import UdpSession
from whatap.net.packet_type_enum import PacketTypeEnum
from whatap.util.date_util import DateUtil

def parseCallbackName(args):
    if args:
        receiver = args[0]
        callbackspan = []

        if len(args) > 2:
            message = args[2]
            if hasattr(message, 'delivery_info'):
                if 'exchange' in message.delivery_info:
                    callbackspan.append(message.delivery_info['exchange'])
                if 'routing_key' in message.delivery_info:
                    callbackspan.append(message.delivery_info['routing_key'])

        if not callbackspan:
            callback = receiver.callbacks[0]
            if isinstance(callback, partial):
                callback = callback.func

            if hasattr(callback, '__qualname__'):
                callbackspan.append(callback.__qualname__)
            else:
                if hasattr(callback, '__module__'):
                    callbackspan.append(callback.__module__)
                if hasattr(callback, '__self__'):
                    callbackspan.append(callback.__self__.__class__.__name__)
                if hasattr(callback, '__name__'):
                    callbackspan.append(callback.__name__)

        return '.'.join(callbackspan)
    return ''

def intercept_receive(fn, *args, **kwargs):
    ctx = TraceContext()
    ctx.service_name = parseCallbackName(args)
    start_interceptor(ctx)

    try:
        callback = fn(*args, **kwargs)
        ctx = TraceContextManager.getLocalContext()
        return callback
    except Exception as e:
        interceptor_step_error(e)
        raise e
    finally:
        if ctx:
            end_interceptor(ctx=ctx)

def parseConnection(conn):
    connstr = ''
    if conn.host:
        connstr += conn.host
    if conn.virtual_host:
        connstr += "/" + conn.virtual_host
    return connstr

def intercept_publish(fn, *args, **kwargs):
    ctx = TraceContextManager.getLocalContext()

    start_time = DateUtil.nowSystem()
    ctx.start_time = start_time

    text = 'rabbitmq://'
    text += parseConnection(args[0].channel.connection)
    text += '?exchage='
    text += args[10]
    text += '&routing='
    text += args[7]
    ctx.active_dbc = text
    ctx.lctx['dbc'] = text

    ctx.active_dbc = 0
    ctx.db_opening = True
    ctx.elapsed = DateUtil.nowSystem() - start_time

    datas = [' ', ' ', 'MQ SESSION INFO: ' + text]
    UdpSession.send_packet(PacketTypeEnum.TX_MSG, ctx, datas)


    start_time = DateUtil.nowSystem()
    ctx.start_time = start_time

    try:
        callback = fn( *args, **kwargs)
        return callback
    except Exception as e:
        interceptor_step_error(e)
    finally:
        ctx.db_opening = False

    return callback

def instrument_kombu(module):
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            callback = intercept_publish(fn, *args, **kwargs)
            return callback
        
        return trace

    if hasattr(module, 'Producer') and hasattr(module.Producer, '_publish'):
        module.Producer._publish = wrapper(module.Producer._publish)

    def wrapper(fn):
        @trace_handler(fn, start=True)
        def trace(*args, **kwargs):
            callback = intercept_receive(fn, *args, **kwargs)
            return callback

        return trace

    if hasattr(module, 'Consumer') and hasattr(module.Consumer, 'receive'):
        module.Consumer.receive = wrapper(module.Consumer.receive)




import logging as logging_module
import os
import platform
import sys
import subprocess
import time
from whatap import build
from whatap.util.date_util import DateUtil
import threading

__version__ = build.version

LOGGING_MSG_FORMAT = '[%(asctime)s] : - %(message)s'
LOGGING_DATE_FORMAT = '%Y-%m-%d  %H:%M:%S'

logging = logging_module.getLogger(__name__)

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

ROOT_DIR = __file__


class ContextFilter(logging_module.Filter):
    def __init__(self):
        super(ContextFilter, self).__init__()
        self.last_id = None
    
    def filter(self, record):
        try:
            if record.id:
                if self.last_id == record.id:
                    return False
                
                self.last_id = record.id
                return True
        
        except Exception as e:
            record.id = ''
            return True


from whatap.conf.configure import Configure as conf

CONFIG_FILE_NAME = 'whatap.conf'
LOG_FILE_NAME = 'whatap-hook.log'


class Logger(object):
    def __init__(self):
        self.logger = logging
        self.logger.addFilter(ContextFilter())
        self.handler = None
        
        self.create_log()
    
    def create_log(self):
        os.environ['WHATAP_LOGS'] = os.path.join(os.environ['WHATAP_HOME'],
                                                 'logs')
        if not os.path.exists(os.environ['WHATAP_LOGS']):
            try:
                os.mkdir(os.environ['WHATAP_LOGS'])
            
            except Exception as e:
                print('WHATAP: LOG FILE WRITE ERROR.')
                print(
                    'WHATAP: Try to execute command. \n  {}'.format(
                        'sudo mkdir -m 777 -p $WHATAP_HOME/logs`'))
        
        self.print_log()
    
    def print_log(self):
        try:
            if self.handler:
                self.logger.removeHandler(self.handler)
            
            temp_logging_msg_format = '[%(asctime)s] : %(id)s - %(message)s'
            logging_format = logging_module.Formatter(
                fmt=temp_logging_msg_format, datefmt=LOGGING_DATE_FORMAT)
            
            fh = logging_module.FileHandler(
                os.path.join(os.environ['WHATAP_LOGS'], LOG_FILE_NAME))
            fh.setFormatter(logging_format)
            self.logger.addHandler(fh)
            self.handler = fh
            
            self.logger.setLevel(logging_module.DEBUG)
        except Exception as e:
            print('WHATAP: LOGGING ERROR: {}'.format(e))
        else:
            self.print_whatap()
    
    def print_whatap(self):
        str = '\n' + \
              ' _      ____       ______' + build.app + '-AGENT  \n' + \
              '| | /| / / /  ___ /_  __/__ ____' + '\n' + \
              '| |/ |/ / _ \\/ _ `// / / _ `/ _ \\' + '\n' + \
              '|__/|__/_//_/\\_,_//_/  \\_,_/ .__/' + '\n' + \
              '                          /_/' + '\n' + \
              'Just Tap, Always Monitoring' + '\n' + \
              'WhaTap ' + build.app + ' Agent version ' + build.version + ', ' + build.release_date + '\n\n'
        
        str += '{0}: {1}\n'.format('WHATAP_HOME', os.environ['WHATAP_HOME'])
        str += '{0}: {1}\n'.format('Config', os.path.join(os.environ['WHATAP_HOME'],
                                                os.environ['WHATAP_CONFIG']))
        str += '{0}: {1}\n\n'.format('Logs', os.environ['WHATAP_LOGS'])
        
        print(str)
        logging.debug(str)


def read_file(home, file_name):
    data = ''
    try:
        f = open(os.path.join(os.environ.get(home), file_name), 'r+')
        data = f.readline()
        f.close()
    finally:
        return data


def write_file(home, file_name, value):
    try:
        f = open(os.path.join(os.environ.get(home), file_name), 'w+')
        f.write(value)
        f.close()
    except Exception as e:
        print('WHATAP: WHATAP HOME ERROR. (path: {})'.format(os.path.join(os.environ.get(home))))
        print(
            'WHATAP: Try to execute command. \n  {}'.format(
                '`sudo chmod -R 777 $WHATAP_HOME`'))
        return False
    else:
        return True


def check_whatap_home(target='WHATAP_HOME'):
    whatap_home = os.environ.get(target)
    if not whatap_home:
        print('WHATAP: ${} is empty'.format(target))
    
    return whatap_home


def init_config(home):
    whatap_home = os.environ.get(home)
    if not whatap_home:
        whatap_home = read_file(home, home.lower())
        if not whatap_home:
            whatap_home = os.getcwd()
            os.environ[home] = whatap_home
            
            print('WHATAP: WHATAP_HOME is empty')
            print(
                'WHATAP: WHATAP_HOME set default CURRENT_WORKING_DIRECTORY value')
            print('CURRENT_WORKING_DIRECTORY is {}\n'.format(whatap_home))
    
    if not write_file(home, home.lower(), whatap_home):
        return False
    
    os.environ[home] = whatap_home
    config_file = os.path.join(os.environ[home],
                               CONFIG_FILE_NAME)
    
    if not os.path.exists(config_file):
        with open(
                os.path.join(os.path.dirname(__file__),
                             CONFIG_FILE_NAME),
                'r') as f:
            content = f.read()
            try:
                with open(config_file, 'w+') as new_f:
                    new_f.write(content)
            except Exception as e:
                print('WHATAP: PERMISSION ERROR: {}'.format(e))
                print(
                    'WHATAP: Try to execute command. \n  {}'.format(
                        '`sudo chmod -R 777 $WHATAP_HOME`'))
                return False
    
    return True


def update_config(home, opt_key, opt_value):
    config_file = os.path.join(os.environ[home],
                               CONFIG_FILE_NAME)
    
    try:
        with open(config_file, 'r+') as f:
            is_update = False
            content = ''
            for line in f:
                if line:
                    key = line.split('=')[0].strip()
                    if key == opt_key:
                        is_update = True
                        line = '{0}={1}\n'.format(key, opt_value)
                    
                    content += line
            if not is_update:
                content += '{0}={1}\n'.format(opt_key, opt_value)
            open(config_file, 'w+').write(content)
    
    except Exception as e:
        print('WHATAP: OPTION ERROR: {}'.format(e))


def config(home):
    os.environ['WHATAP_CONFIG'] = CONFIG_FILE_NAME
    
    from whatap.conf.configure import Configure as conf
    if conf.init():
        from whatap.net.packet_enum import PacketEnum
        PacketEnum.PORT = int(conf.net_udp_port)

        from whatap.conf.license import License
        conf.PCODE = License.getProjectCode(conf.license)

        hooks(home)


from whatap.trace.trace_import import ImportFinder
from whatap.trace.trace_module_definition import DEFINITION, IMPORT_HOOKS, \
    PLUGIN


def hooks(home):
    try:
        for key, value_list in DEFINITION.items():
            for value in value_list:
                if len(value) == 3 and not value[2]:
                    continue
                
                IMPORT_HOOKS[value[0]] = {'def': value[1],
                                          'module': '{0}.{1}.{2}.{3}'.format(
                                              'whatap',
                                              'trace',
                                              'mod',
                                              key)}
    except Exception as e:
        logging.debug(e, extra={'id': 'MODULE ERROR'})
    finally:
        try:
            if conf.hook_method_patterns:
                from whatap.trace.mod.plugin import instrument_plugin
            
                patterns = conf.hook_method_patterns.split(',')
                for pattern in patterns:
                    pattern=pattern.strip()
                
                    module_name, class_def = pattern.split(':')
                    if not PLUGIN.get(module_name):
                        PLUGIN[module_name] = []
                    PLUGIN[module_name].append(class_def)
                    
                    DEFINITION["plugin"].append(
                        (module_name, 'instrument_plugin'))
                    
                    key = 'plugin'
                    for value in DEFINITION[key]:
                        IMPORT_HOOKS[value[0]] = {'def': value[1],
                                                  'module': '{0}.{1}.{2}.{3}'.format(
                                                      'whatap',
                                                      'trace',
                                                      'mod',
                                                      key)}
        
        except Exception as e:
            logging.debug(e, extra={'id': 'PLUGIN ERROR'})
        finally:
            sys.meta_path.insert(0, ImportFinder())
            logging.debug('WHATAP AGENT START!', extra={'id': 'WA000'})


def agent():
    home = 'WHATAP_HOME'
    whatap_home = os.environ.get(home)
    if not whatap_home:
        whatap_home = read_file(home, home.lower())
        if not whatap_home:
            whatap_home = os.getcwd()
            os.environ[home] = whatap_home
            
            print('WHATAP: WHATAP_HOME is empty')
            print(
                'WHATAP: WHATAP_HOME set default CURRENT_WORKING_DIRECTORY value')
            print('CURRENT_WORKING_DIRECTORY is {}\n'.format(whatap_home))
    
    if write_file(home, home.lower(), whatap_home):
        os.environ['WHATAP_HOME'] = whatap_home
        
        go()
        config(home)
        #startWhatapThread()

def startWhatapThread():
    def __whatap_jobs():
        from whatap.conf.configure import Configure as conf
        from whatap.net.udp_session import UdpSession
        UdpSession.read_timeout = conf.session_read_timeout
        UdpSession.udp()

        def __udp_recv():
            UdpSession.get()

        def __config_load():
            now = time.time()
            if now - conf.last_loaded > conf.init_load_interval / 1000:
                conf.init(False)

        while True:
            for job in (__udp_recv, __config_load):
                try:
                    job()
                    time.sleep(0)
                except Exception as e:
                    logging.debug(e, extra={'id': 'WA914'}, exc_info=True)
                    time.sleep(0.05)

    t = threading.Thread(target=__whatap_jobs)
    t.setDaemon(True)
    t.start()

ARCH = {
    'x86_64': 'amd64',
    'x86': '386',
    'x86_32': '386',
    'ARM': 'arm',
    'AArch64': 'arm64',
}

AGENT_NAME = 'whatap_python'


def go(batch=False):
    os.environ['WHATAP_VERSION'] = build.version
    
    os.environ['whatap.start'] = str(DateUtil.now())
    
    os.environ['python.uptime'] = str(DateUtil.datetime())
    os.environ['python.version'] = sys.version
    os.environ['python.tzname'] = time.tzname[0]

    
    os.environ['os.release'] = platform.release()
    
    os.environ['sys.version_info'] = str(sys.version_info)
    os.environ['sys.executable'] = sys.executable
    os.environ['sys.path'] = str(sys.path)
    
    if not batch:
        home = 'WHATAP_HOME'
        file_name = AGENT_NAME + '.pid'
    else:
        home = 'WHATAP_HOME_BATCH'
        file_name = AGENT_NAME + '.pid.batch'
    
    if go_kill(home, file_name, False):
        try:
            agent_cwd = os.environ.get(home)
            if os.path.exists(os.path.join(agent_cwd, AGENT_NAME)):
                os.remove(os.path.join(agent_cwd, AGENT_NAME))

            source_cwd = os.path.join(os.path.join(os.path.dirname(__file__), 'agent'), platform.system().lower(),
                                      ARCH[platform.machine()])
            
                
            os.symlink(os.path.join(source_cwd, AGENT_NAME),
                       os.path.join(agent_cwd, AGENT_NAME))
            
            os.environ['whatap.enabled'] = 'True'
            process = subprocess.Popen('./{0}'.format(AGENT_NAME),
                                       cwd=agent_cwd)
            
            # Write PID file
            write_file(home, file_name, str(process.pid))
        
        except Exception as e:
            print('WHATAP: AGENT ERROR: {}'.format(e))
        else:
            print('WHATAP: AGENT UP! (process name: {})\n'.format(AGENT_NAME))


import signal


def go_kill(home, file_name, is_kill=True):
    if is_kill == False:
        return True

    try:
        pids = subprocess.check_output('pgrep {0}'.format(AGENT_NAME),
                                       shell=True).split()
    except Exception as e:
        return True
    else:
        pid = read_file(home, file_name)
        try:
            if len(pid.split()):
                os.kill(int(pid), signal.SIGKILL)
            else:
                if not read_file('WHATAP_HOME', AGENT_NAME + '.pid') \
                        and not read_file('WHATAP_HOME_BATCH',
                                          AGENT_NAME + '.pid.batch'):
                    for pid in pids:
                        os.kill(int(pid), signal.SIGKILL)
                    
                    write_file(home, AGENT_NAME + '.pid', '')
                    write_file(home, AGENT_NAME + '.pid.batch', '')
        
        
        except Exception as e:
            logging.debug(e, extra={'id': ' AGENT UPGRADE ERROR'})
            logging.debug(
                'Try to execute command. \n  {}'.format(
                    '`ps -ef | grep {}`'.format(AGENT_NAME)))
            logging.debug(
                'Next, `sudo kill -9 {}` or `sudo killall {}`'.format(
                    pid, AGENT_NAME))
            return False
        else:
            print('WHATAP: AGENT DOWN..')
            return True


from whatap.trace.mod.application_wsgi import interceptor, start_interceptor, \
    end_interceptor, trace_handler

from whatap.trace.trace_context import TraceContext

def register_app(fn):
    @trace_handler(fn, True)
    def trace(*args, **kwargs):
        callback = None
        try:
            environ = args[0]
            callback = interceptor((fn, environ), *args, **kwargs)
        except Exception as e:
            logging.debug('WHATAP(@register_app): ' + str(e),
                          extra={'id': 'WA777'}, exc_info=True)
        finally:
            return callback if callback else fn(*args, **kwargs)
    
    if not os.environ.get('whatap.enabled'):
        agent()
    
    return trace


def method_profiling(fn):
    def trace(*args, **kwargs):
        callback = None
        try:
            ctx = TraceContext()
            ctx.service_name=fn.__name__
            ctx = start_interceptor(ctx)
            callback = fn(*args, **kwargs)
            end_interceptor(thread_id= ctx.thread_id)
        except Exception as e:
            logging.debug('WHATAP(@method_profiling): ' + str(e),
                          extra={'id': 'WA776'}, exc_info=True)
        finally:
            return callback if callback else fn(*args, **kwargs)
    
    if not os.environ.get('whatap.enabled'):
        agent()
    
    return trace


def batch_agent():
    home = 'WHATAP_HOME_BATCH'
    batch_home = os.environ.get(home)
    if not batch_home:
        if not read_file(home, home.lower()):
            print('WHATAP: WHATAP_HOME_BATCH is empty')
            return
    
    if write_file(home, home.lower(), batch_home):
        os.environ['WHATAP_HOME_BATCH'] = batch_home
        os.environ['WHATAP_HOME'] = batch_home
        go(batch=True)


def batch_profiling(fn):
    import inspect
    frame = inspect.stack()[1]
    module = inspect.getmodule(frame[0])
    
    def trace(*args, **kwargs):
        if not os.environ.get('whatap.batch.enabled'):
            home = 'WHATAP_HOME_BATCH'
            batch_home = read_file(home, home.lower())
            if not batch_home:
                print(
                    'WHATAP(@batch_profiling): try, whatap-start-batch')
                return fn(*args, **kwargs)
            else:
                os.environ['whatap.batch.enabled'] = 'True'
                os.environ['WHATAP_HOME_BATCH'] = batch_home
                os.environ['WHATAP_HOME'] = batch_home
                config(home)
        
        callback = None
        try:
            ctx = TraceContext()
            ctx.service_name=module.__file__.split('/').pop()
            ctx = start_interceptor(ctx)
            
            callback = fn(*args, **kwargs)
            end_interceptor(thread_id=ctx.thread_id)
        except Exception as e:
            logging.debug('WHATAP(@batch_profiling): ' + str(e),
                          extra={'id': 'WA777'}, exc_info=True)
        finally:
            return callback if callback else fn(*args, **kwargs)
    
    return trace



Configuration = {
    "dev": False,
    "net_udp_port": "6600",
    "web_static_content_extensions": "js, htm, html, gif, png, jpg, css, swf, ico",
    "trace_auto_normalize_enabled": True,
    "trace_user_enabled": True,
    "trace_user_using_ip": True,

    "query_string_enabled": False,
    "query_string_urls": "",

    "profile_dbc_close": False,
    "profile_http_header_enabled": False,
    "profile_http_parameter_enabled": False,
    "profile_sql_param_enabled": False,
    "hook_method_patterns": "",

    "mtrace_rate": 0,
    "mtrace_spec": '',
    "mtrace_send_url_length": 80,

    "_trace_mtrace_caller_key": "x-wtap-mst",
    "_trace_mtrace_callee_key": "x-wtap-tx",
    "_trace_mtrace_info_key": "x-wtap-inf",
    "_trace_mtrace_ip": "x-wtap-ip",

    "stat_enabled": False,
    "stat_domain_enabled": False,
    "stat_domain_max_count": 7000,
    "stat_mtrace_enabled": False,
    "stat_mtrace_max_count": 7000,
    "stat_login_enabled": False,
    "stat_login_max_count": 7000,
    "stat_referer_enabled": False,
    "stat_referer_max_count": 7000,

	"stat_tx_max_count": 5000,
	"stat_sql_max_count": 5000,
	"stat_httpc_max_count": 5000,
	"stat_error_max_count": 1000,
	"stat_useragent_max_count": 500,

    "user_header_ticket_enabled": False,
    "user_header_ticket":"",
    "trace_user_cookie_limit":2048,
    "trace_user_cookie_domain":"",
    "profile_exception_stack": False,

    # "whatap_server_host": "127.0.0.1/127.0.0.1",
    # "whatap_server_port": 6600,
    # "mode": "whatap_dev",
    # "debug": False,
    # "license": None,
    # "counter_enabled": True,
    # "stat_enabled": True,
    # "active_stack_enabled": True,
    # "trx_profile_enabled": True,
    # "trx_profile_sql_enabled": True,
    # "trx_profile_httpc_enabled": True,
    # "trx_profile_max_time": 300000,
    # "trx_profile_max_count": 10000,
    # "prepared_sql_max": 5001,
    # "debug_prepared_sql": False,
    # "trace_service_name_header_key": None,
    # "trace_service_name_key": None,
    # "web_static_content_extensions": "js, htm, html, gif, png, jpg, css, txt",
    # "trace_auto_service_enabled": False,
    # "trace_auto_service_backstack_enabled": True,
    # "trace_background_socket_enabled": True,
    # "trace_activeserivce_yellow_time": 3000,
    # "trace_activeservice_red_time": 8000,
    # "trace_activeservice_hang_time": 6000,
    # "trace_intertx_enabled": True,
    # "trace_error_callstack_depth": 50,
    # "trace_active_callstack_depth": 50,
    # "_trace_intertx_caller_pcode_key": "x-wtap-pcode",
    # "_trace_intertx_caller_oid_key": "x-wtap-oid",
    # "_trace_intertx_caller_seq_key": "x-wtap-seq",
    # "_trace_intertx_call_url_key": "x-wtap-url",
    # "trace_httpc_normalize_urls": "",
    # "trace_httpc_normalize_enabled": True,
    # "trace_normalize_enabled": True,
    # "log_ignore_set": None,
    # "tcp_connection_timeout": 60000,
    # "tcp_so_timeout": 60000,
    # "net_send_max_bytes": 5 * 1024 * 1024,
    # "net_send_queue1_size": 512,
    # "net_send_queue2_size": 1024,
    # "hook_connection_open_patterns": "",
    # "log_datasource_lookup_enabled": True,
    "throttle_enabled": False,
    "throttle_blocking_url": '',
    "throttle_blocking_ip": '',
    "throttle_blocked_forward": '',
    "throttle_blocked_message": 'request blocked!!',
    "throttle_limit": None,
    "throttle_target_urls": '',
    "throttle_passing_url": '',
    "throttle_passing_url_prefix": '',
    "throttle_rejected_forward": '',
    "throttle_rejected_message": "too many request!!",
    "reject_event_enabled": False,
    # "profile_http_querystring_enabled": False,
    # "profile_http_header_enabled": False,
    # "profile_http_parameter_enabled": False,
    # "profile_http_parameter_url_prefix": "/",
    # "profile_http_header_url_prefix": "/",
    # "profile_connection_open_enabled": True,
    # "profile_step_max_count": 1024,
    # "hook_method_patterns": "",
    # "hook_method_access_public_enabled": True,
    # "hook_method_access_protected_enabled": True,
    # "hook_method_access_private_enabled": False,
    # "hook_method_access_none_enabled": True,
    # "hook_method_ignore_prefixes": "get,set",
    # "hook_method_ignore_classes": "",
    # "profile_method_enabled": True,
    # "hook_service_patterns": "",
    # "hook_httpc_patterns": "",
    # "hook_httpc_info_patterns": "",
    # "hook_jdbc_pstmt_classes": "",
    # "hook_jdbc_stmt_classes": "",
    # "hook_jdbc_rs_classes": "",
    # "hook_jdbc_wrapping_driver_patterns": "",
    # "hook_context_classes": "javax/naming/InitialContext",
    # "hook_jsp_patterns": "",
    # "hook_trace_helper_start_patterns": "",
    # "hook_trace_helper_end_patterns": "",
    # "hook_dyna_ignore_count": 100,
    #
    # "profile_fullstack_max_lines": 0,
    #
    # "profile_step_normal_count": 800,
    # "profile_step_heavy_count": 1000,
    # "profile_step_max_count": 1024,
    # "profile_step_heavy_time": 100,
    #
    # "boolean profile_sql_param_enabled": False,
    # "profile_sql_resource_enabled": False,
    # "profile_method_resource_enabled": False,
    # "profile_httpc_resource_enabled": False,
    # "_profile_prepare_sql": False,
    # "profile_dbc_close": False,
    #
    "trace_http_client_ip_header_key": "",
    # "trace_user_enabled": True,
    # "trace_user_using_ip": False,
    # "trace_user_using_type": 1,
    # "trace_user_cookie_limit": 2048,
    # "profile_error_jdbc_fetch_max": 10000,
    # "profile_error_sql_time_max": 30000,
    # "hook_direct_patch_classes": "",
    # "_hook_serivce_enabled": True,
    # "_hook_dbsql_enabled": True,
    # "_hook_dbconn_enabled": True,
    # "_hook_methods_enabled": True,
    # "_hook_socket_enabled": True,
    # "_hook_usertx_enabled": True,
    # "log_rotation_enabled": True,
    # "log_keep_days": 7,
    # "thread_dump_enabled": False,
    # "debug_cpu_enabled": False,
    # "active_stack_always_do": False,
    # "active_stack_callstack_interval_sec": 10,
    # "realtime_interval_ms": 5000,
    # "cipher_net_enabled": True,
    # "cipher_level": 2,
    # 'cypher_level': 128,
    # "boot_redefine_size": 100,
    # "counter_netstat_enabled": False,
    # "counter_procfd_enabled": False,
    # "counter_interval": 5000,
    # "CONFIG_INITED": False,
    # "_profile_position_httpc_hash": 0,
    # "profile_position_httpc": "",
    # "_profile_position_sql_hash": 0,
    # "profile_position_sql": "",
    # "_profile_position_method_hash": 0,
    # "profile_position_method": "",
    # "profile_position_depth": 50,
    # "time_sync_interval_ms": 300000,
    # "trace_active_transaction_yellow_time": 3000,
    # "trace_active_transaction_red_time": 8000,
    # "trace_active_transaction_hang_time": 30000,
    # "realtime_user_thinktime_max": 300000,
    # "service_port_enabled": False,
    # 'log_pack_data': False,
    # "active_stack_second": 10
    "init_load_interval": 30000,
    "session_read_timeout": 5000,
    "reject_event_interval": 30000,
    "master_agent_host" : "whatap-master-agent.whatap-monitoring.svc.cluster.local",
    "master_agent_port" : 6600,
}



import threading, time
import logging as logging_module
logging = logging_module.getLogger(__name__)

def startWhatapThread():
    def __whatap_jobs():
        from whatap.conf.configure import Configure as conf
        from whatap.net.udp_session import UdpSession
        UdpSession.read_timeout = conf.session_read_timeout
        UdpSession.udp()

        def __udp_recv():
            UdpSession.get()

        def __config_load():
            now = time.time()
            if now - conf.last_loaded > conf.init_load_interval / 1000:
                conf.init(False)

        while True:
            for job in (__udp_recv, __config_load):
                try:
                    job()
                    time.sleep(0)
                except Exception as e:
                    logging.debug(e, extra={'id': 'WA914'}, exc_info=True)
                    time.sleep(0.05)

    t = threading.Thread(target=__whatap_jobs)
    t.setDaemon(True)
    t.start()

startWhatapThread()


from whatap.net.packet_type_enum import PacketTypeEnum
from whatap.net.udp_session import UdpSession
from whatap.trace.mod.application_wsgi import trace_handler, interceptor_db_execute
from whatap.util.date_util import DateUtil
from whatap.trace.trace_context_manager import TraceContextManager

def instrument_sqlalchemy(module):
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            ctx = TraceContextManager.getLocalContext()
            
            start_time = DateUtil.nowSystem()
            ctx.start_time = start_time
            
            text = args[0].bind.url.__to_string__().replace('***', '')
            ctx.lctx['dbc'] = text
            
            datas = [' ', ' ', 'DB SESSION INFO: ' + text]
            ctx.elapsed = DateUtil.nowSystem() - start_time
            UdpSession.send_packet(PacketTypeEnum.TX_MSG, ctx, datas)
            
            callback = fn(*args, **kwargs)
            return callback
        
        return trace
    
    module.Session.get_bind = wrapper(module.Session.get_bind)

def addQuote(arg_dict):
    quoted_dict = dict()

    for k, v in arg_dict.items():
        if isinstance(v, str):
            quoted_dict[k] = "'" + v.replace("'", "\\'") + "'"
        else:
            quoted_dict[k] = v

    return quoted_dict

def instrument_sqlalchemy_engine(module):
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            # print('instrument_sqlalchemy_engine step -1')
            ctx = TraceContextManager.getLocalContext()
            # print('instrument_sqlalchemy_engine CTX:', ctx)
            cursor = args[1]
            query = None
            if len(args) > 3 and type(args[3]) == dict and args[3]:
                try:
                    query = args[2] % addQuote(args[3])
                except Exception as e:
                    pass
            # print('instrument_sqlalchemy_engine 2:', query)
            try:
                if not query:
                    query = args[2].decode()
            except Exception as e:
                query = str(args[2])
            # print('instrument_sqlalchemy_engine 3:', query)
            if not query or ctx.active_sqlhash:
                return fn(*args, **kwargs)

            start_time = DateUtil.nowSystem()
            ctx.start_time = start_time
            ctx.active_sqlhash = query
            # print('instrument_sqlalchemy_engine 4')
            try:
                callback = fn(*args, **kwargs)
                # print('instrument_sqlalchemy_engine 5')
                return callback
            except Exception as e:
                interceptor_step_error(e)
            finally:
                # print('instrument_sqlalchemy_engine 6')
                datas = [ctx.lctx.get('dbc', ''), query]
                ctx.elapsed = DateUtil.nowSystem() - start_time
                # print('instrument_sqlalchemy_engine 6.1', datas)
                UdpSession.send_packet(PacketTypeEnum.TX_SQL, ctx,
                                       datas)
                # print('instrument_sqlalchemy_engine 7')
                count = cursor.rowcount
                # print('instrument_sqlalchemy_engine 8')
                if count > -1:
                    desc = '{0}: {1}'.format('Fetch count', count)
                    datas = [' ', ' ', desc]
                    ctx.elapsed = 0
                    UdpSession.send_packet(PacketTypeEnum.TX_MSG, ctx, datas)
                ctx.active_sqlhash = 0
                # print('instrument_sqlalchemy_engine 9')

        return trace

    # print('database_toolkit.instrument_sqlalchemy_engine step -1')
    if hasattr(module, 'DefaultDialect') and hasattr(module.DefaultDialect, 'do_execute'):
        # print('database_toolkit.instrument_sqlalchemy_engine step -2')
        module.DefaultDialect.do_execute = wrapper(module.DefaultDialect.do_execute)
        module.DefaultDialect.do_executemany = wrapper(module.DefaultDialect.do_executemany)
        module.DefaultDialect.do_execute_no_params = wrapper(module.DefaultDialect.do_execute_no_params)
        # print('database_toolkit.instrument_sqlalchemy_engine step -3')


class PacketTypeEnum(object):
    TX_BLANK = 0
    TX_START = 1
    TX_DB_CONN = 2
    TX_DB_FETCH = 3
    TX_SQL = 4
    TX_SQL_START = 5
    TX_SQL_END = 6

    TX_HTTPC = 7
    TX_HTTPC_START = 8
    TX_HTTPC_END = 9

    TX_ERROR = 10
    TX_MSG = 11
    TX_METHOD = 12

    # secure msg
    TX_SECURE_MSG = 13

    TX_END = 255

    TX_PARAM = 30
    ACTIVE_STACK = 40
    ACTIVE_STATS = 41

    EVENT = 50



from whatap.io.data_inputx import DataInputX
from whatap.io.data_outputx import DataOutputX


class IPUtil(object):
    empty = bytearray([0, 0, 0, 0])

    @classmethod
    def toString(cls, ip):
        if isinstance(ip, int):
            return cls.toString(DataOutputX.toBytes(ip))
        elif isinstance(ip, bytes) or isinstance(ip, bytearray):
            try:
                to_str = '{0}.{1}.{2}.{3}'.format(ip[0] & 0xff, ip[1] & 0xff,
                                                  ip[2] & 0xff, ip[3] & 0xff)
                return to_str
            except Exception:
                return '0.0.0.0'

    @classmethod
    def toBytes(cls, ip):
        if isinstance(ip, int):
            return DataOutputX.toBytes(ip)
        elif isinstance(ip, str):
            # if not ip:
            #     return cls.empty

            try:
                ss = ip.split('.')
                ss_ln = len(ss)
                if ss_ln != 4:
                    return cls.empty

                result = bytearray(ss_ln)
                for i in range(ss_ln):
                    s = int(ss[i])
                    if s < 0 or s > 0xff:
                        return cls.empty

                    result[i] = s & 0xff
                return result
            except Exception as e:
                import traceback
                print('iputil toBytes ',e)
                traceback.print_stack()
                return cls.empty

    @classmethod
    def toInt(cls, ip):
        if isinstance(ip, bytes):
            if cls.isOK(ip):
                return DataInputX.toInt(ip, 0)
            else:
                return 0
        elif isinstance(ip, str):
            return DataInputX.toInt(cls.toBytes(ip), 0)

    @classmethod
    def isOK(cls, ip):
        return ip is not None and len(ip) == 4

    @classmethod
    def isNotLocal(cls, ip):
        return cls.isOK(ip) and (ip[0] & 0xff) != 127



import inspect
from whatap.net.packet_type_enum import PacketTypeEnum
from whatap.net.udp_session import UdpSession
from whatap.trace import get_dict
from whatap.trace.mod.application_wsgi import trace_handler
from whatap.util.date_util import DateUtil
from whatap.trace.trace_context_manager import TraceContextManager

from whatap import logging


def instrument_plugin(module_dict):
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            module_name = fn.__module__
            def_name = fn.__name__
            class_name = args[0].__class__.__name__ if len(args) and hasattr(
                args[0], def_name) else None
            
            if class_name:
                text = '{}:{}.{}'.format(module_name, class_name, def_name)
            else:
                text = '{}:{}'.format(module_name, def_name)
            
            ctx = TraceContextManager.getLocalContext()
            start_time = DateUtil.nowSystem()
            ctx.start_time = start_time
            try:
                callback = fn(*args, **kwargs)
                return callback
            except Exception as e:
                logging.debug(e, extra={'id': 'hook_method_patterns PARSING ERROR'})
            finally:
                datas = [text, '']
                ctx.elapsed = DateUtil.nowSystem() - start_time
                UdpSession.send_packet(PacketTypeEnum.TX_METHOD, ctx, datas)
        
        return trace
    
    try:
        for class_def in module_dict['class_defs']:
            class_def_list = class_def.split('.')  # exception
            
            m_list = []
            t_list = []
            module = module_dict['module']
            if len(class_def_list) == 2:
                if class_def_list[0] == '*':
                    for attr in dir(module):
                        if not attr.startswith('__') \
                                and not attr.endswith('__') \
                                and inspect.isclass(getattr(module, attr)) \
                                and not isinstance(getattr(module, attr), type(module)) \
                                and module.__name__ == getattr(module, attr).__module__ :
                            m_list.append(getattr(module, attr))
                else:
                    m_list = [getattr(module, class_def_list[0])]
                    
                target = class_def_list[1]
            else:
                m_list = [module]
                target = class_def_list[0]
                
            for m in m_list:
                if target == '*':
                    for attr in dir(m):
                        if not attr.startswith('__') \
                                and not attr.endswith('__') \
                                and hasattr(m, attr) \
                                and inspect.isfunction(getattr(m, attr)):
                            t_list.append(attr)
                else:
                    t_list.append(target)
                
                for t in t_list:
                    if hasattr(m, t):
                        try:
                            setattr(m, t, wrapper(getattr(m, t)))
                        except Exception as e:
                            get_dict(m)[t] = wrapper(getattr(m, t))
    except Exception as e:
        logging.debug(e, extra={'id': 'hook_method_patterns PARSING ERROR'})
        



from whatap.util.date_util import DateUtil
import random
try:
    random.seed(DateUtil.nowSystem())
except:
    pass

class KeyGen(object):
    @classmethod
    def next(cls):
        return int(random.random()*0x7fffffff)



from whatap.trace import get_dict
from whatap.trace.mod.application_wsgi import trace_handler, \
    interceptor_step_error, start_interceptor, end_interceptor
from whatap.trace.trace_context import TraceContext
from whatap.trace.trace_context_manager import TraceContextManager
from whatap.net.udp_session import UdpSession
from whatap.net.packet_type_enum import PacketTypeEnum
from whatap.util.date_util import DateUtil

def intercept_worker(fn, *args, **kwargs):
    ctx = TraceContext()
    worker_ctx = args[1]
    ctx.service_name = worker_ctx.call_id_stack
    start_interceptor(ctx)

    try:
        callback = fn(*args, **kwargs)
        ctx = TraceContextManager.getLocalContext()
        return callback
    except Exception as e:
        interceptor_step_error(e)
        raise e
    finally:
        if ctx:
            end_interceptor(ctx=ctx)

def instrument_nameko_spawn_worker(module):
    def wrapper(fn):
        @trace_handler(fn, start=True)
        def trace(*args, **kwargs):
            callback = intercept_worker(fn, *args, **kwargs)
            return callback

        return trace

    if hasattr(module, 'ServiceContainer') and hasattr(module.ServiceContainer, '_run_worker'):
        module.ServiceContainer._run_worker = wrapper(module.ServiceContainer._run_worker)




#!/usr/bin/env python
import os

from datetime import date

from setuptools import setup, find_packages

from whatap import build
build.release_date = date.today().strftime('%Y%m%d')

readme_file = os.path.join(os.getcwd(), 'whatap', 'README.rst')

setup(name=build.name,
      version=build.version,
      description='Monitoring and Profiling Service',
      long_description=open(readme_file).read(),
      author='whatap',
      author_email='admin@whatap.io',
      license='Whatap License',
      url='https://www.whatap.io',
      packages=find_packages(exclude=('whatap_dev', 'whatap_dev.*')),
      package_data={
          'whatap': ['LICENSE', '*.rst', '*.conf', '*.json', 'agent/*/*/whatap_python']
      },
      entry_points={
          'console_scripts': [
              'whatap-start-agent=whatap.scripts:start_agent',
              'whatap-stop-agent=whatap.scripts:stop_agent',
              'whatap-setting-config=whatap.scripts:setting_config',
              
              
              'whatap-start-batch-agent=whatap:start_batch_agent',
          ],
      },
      zip_safe=False,
      )




import sys

try:
    from importlib import find_loader
except ImportError:
    find_loader = None

from whatap.trace.trace_module_definition import IMPORT_HOOKS, PLUGIN

from whatap import logging
from whatap.conf.configure import Configure as conf

def load_module(module, fullname):
    # if conf.dev:
    #     logging.debug(fullname)

    try:
        if fullname in IMPORT_HOOKS:
            wfullname = IMPORT_HOOKS[fullname]['module']
            wmodule = sys.modules.get(wfullname)
            if not wmodule:
                __import__(wfullname)
                wmodule = sys.modules.get(wfullname)

            if wfullname.endswith('plugin'):
                module = {'module': module, 'class_defs': PLUGIN[fullname]}
                
            getattr(wmodule,
                    IMPORT_HOOKS[fullname]['def'])(module)

    except Exception as e:
        logging.debug(e, exc_info=True)
    finally:
        return module

class _ImportHookLoader(object):
    def load_module(self, fullname):
        module = sys.modules[fullname]
        return load_module(module, fullname)

class _ImportHookChainedLoader(object):
    def __init__(self, loader):
        self.loader = loader
    
    def load_module(self, fullname):
        module = self.loader.load_module(fullname)
        return load_module(module, fullname)

class ImportFinder(object):
    def __init__(self):
        self._hooks = {}
    
    def find_module(self, fullname, path=None):
        if fullname not in IMPORT_HOOKS \
                or fullname.startswith('whatap') \
                or fullname.startswith('pip'):
            return None

        if fullname in self._hooks:
            return None
        self._hooks[fullname] = True
        
        # if conf.dev:
        #     logging.debug(fullname)
        
        try:
            if find_loader:
                loader = find_loader(fullname, path)
                
                if loader:
                    return _ImportHookChainedLoader(loader)
            else:
                __import__(fullname)
                return _ImportHookLoader()
        
        except Exception as e:
            return



from whatap.util.hash_util import HashUtil as hash_util
from whatap.util.hexa32 import Hexa32 as hexa32
from whatap.util.keygen import KeyGen
import whatap.util.bit_util as bit_util
from whatap.conf.configure import Configure as conf
import logging as logging_module
import sys,  traceback

logger = logging_module.getLogger(__name__)

def toDjangoHeaderName(src):

    return 'HTTP_' + src.upper().replace('-','_')

class UseridUtil(object):
    WHATAP_R = "WHATAP"

    @staticmethod
    def getUserId(req,  defValue):
        try:

            if conf.user_header_ticket:
                ticket = UseridUtil.getHeader(req, conf.user_header_ticket)
                if  ticket:
                    return hash_util.hashFromString(ticket), ticket
                return 0,""
            cookie = UseridUtil.getHeader(req, "Cookie")
            if cookie:
                if len(cookie) >= conf.trace_user_cookie_limit :
                    return defValue

                x1 = cookie.find(UseridUtil.WHATAP_R)
                if x1 >= 0:
                    x2 = cookie.find(';', x1)
                    if x2 > 0:
                        value = cookie[x1 + len(UseridUtil.WHATAP_R) + 1: x2]
                    else:
                        value = cookie[x1 + len(UseridUtil.WHATAP_R) + 1:]
                    return hexa32.toLong32(value), value
            userid=KeyGen.next()
            return userid, hexa32.toString32(userid)
        except Exception:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            logger.debug("A502",10, str(exc_value))

    @staticmethod
    def setUserId(req, res, cookieValue):
        try:
            if not conf.user_header_ticket:
                cookie = UseridUtil.getHeaderEx(req, "Cookie")
                if not cookie:
                    UseridUtil.setCookie(res, UseridUtil.WHATAP_R, cookieValue, bit_util.INT_MAX_VALUE, "/", conf.trace_user_cookie_domain )
        except Exception:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            logger.debug("A503", 10, str(exc_value))

    @staticmethod
    def DEPRECATED_getUserid(req, res, defValue):
        try:
            if conf.user_header_ticket_enabled:
                ticket = UseridUtil.getHeader(req, conf.user_header_ticket)
                if  ticket:
                    return hash_util.hash(ticket)
                return 0
            cookie = UseridUtil.getHeader(req, "Cookie")
            if cookie:
                if len(cookie) >= conf.trace_user_cookie_limit :
                    return defValue

                x1 = cookie.find(UseridUtil.WHATAP_R)
                if x1 >= 0:
                    x2 = cookie.find(';', x1)
                    if x2 > 0:
                        value = cookie[x1 + len(UseridUtil.WHATAP_R) + 1: x2]
                    else:
                        value = cookie[x1 + len(UseridUtil.WHATAP_R) + 1:]
                    return hexa32.toLong32(value)
            UseridUtil.setCookie(res, UseridUtil.WHATAP_R, hexa32.toString32(KeyGen.next()), bit_util.INT_MAX_VALUE, "/", conf.trace_user_cookie_domain )

        except Exception:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            logger.debug("A502",10, str(exc_value))

        return defValue

    META_DICT = {'django': lambda x, name: x.get(toDjangoHeaderName(name))}

    @staticmethod
    def getHeader(args, name):
        for arg in args:
            if isinstance(arg, dict):
                if name in arg:
                    return arg[name]
                else:
                    for  fn in UseridUtil.META_DICT.values():
                        v = fn(arg, name)
                        if v:
                            return v

        return None

    GET_HEADER={'META': lambda x : x.META.get}
    @staticmethod
    def getHeaderEx(req, name):
        name = toDjangoHeaderName(name)

        for k, fn in UseridUtil.GET_HEADER.items():
            if hasattr(req, k):
                return fn(req)(name)

        return None

    SETCOOKIE_METHODS={
       "set_cookie": lambda res, key, value, max_age, path, domain : res.set_cookie(key, value, max_age = max_age, path=path, domain=domain )
    }
    @staticmethod
    def setCookie(res, key, value, max_age, path, domain):
        for mname, fn in UseridUtil.SETCOOKIE_METHODS.items():
            if hasattr(res, mname):
                fn(res, key, value, max_age, path, domain)
                break

    @staticmethod
    def getRemoteAddr(args):
        if conf.trace_http_client_ip_header_key:
            header_val = UseridUtil.getHeader(args, conf.trace_http_client_ip_header_key)
            if header_val:
                return header_val.split(',')[0].strip()

        return UseridUtil.getHeader(args, "REMOTE_ADDR")


from whatap.trace.mod.application_wsgi import interceptor, trace_handler, \
    interceptor_error


def instrument(module):
    def wrapper(fn):
        @trace_handler(fn, True)
        def trace(*args, **kwargs):
            callback = interceptor(fn, *args, **kwargs)
            return callback
        
        return trace
    
    module.Bottle.__call__ = wrapper(module.Bottle.__call__)
    
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            callback = fn(*args, **kwargs)
            if len(args) > 3:
                e = args[3]
                status_code = args[1]
            else:
                e = args[0]
                status_code = e._status_code
            
            errors = [e.__class__.__name__, str(e)]
            interceptor_error(status_code, errors)
            return callback
        
        return trace
    
    module.HTTPError.__init__ = wrapper(module.HTTPError.__init__)



from whatap.trace.mod.application_wsgi import transfer, trace_handler, \
    interceptor_httpc_request


def instrument_requests(module):
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            # set mtid header
            args[1].headers = transfer(args[1].headers)
            
            # set httpc_url
            httpc_url = args[1].url
            callback = interceptor_httpc_request(fn, httpc_url, *args, **kwargs)
            return callback
        
        return trace
    
    module.Session.send = wrapper(module.Session.send)



from whatap.trace.mod.application_wsgi import transfer, interceptor_httpc_request, \
    trace_handler, interceptor_sock_connect


def instrument_urllib3(module):
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            # set mtid header
            kwargs['headers'] = transfer(kwargs.get('headers', {}))
            
            # set httpc_url
            httpc_url = args[2]
            callback = interceptor_httpc_request(fn, httpc_url, *args, **kwargs)
            return callback
        
        return trace
    if hasattr(module, 'RequestMethods'):
        module.RequestMethods.request = wrapper(module.RequestMethods.request)
    if hasattr(module, 'HTTPConnection'):
        module.HTTPConnection.connect = wrapper(interceptor_sock_connect)


from whatap.trace.mod.application_wsgi import interceptor, trace_handler, \
    interceptor_error


def instrument(module):
    def wrapper(fn):
        @trace_handler(fn, True)
        def trace(*args, **kwargs):
            callback = interceptor(fn, *args, **kwargs)
            return callback
        
        return trace
    
    module.Application.__call__ = wrapper(module.Application.__call__)
    
    def wrapper(fn):
        @trace_handler(fn)
        def trace(*args, **kwargs):
            callback = fn(*args, **kwargs)
            
            e = args[0]
            status_code = e.code
            errors = [e.reason, e._message]
            interceptor_error(status_code, errors)
            return callback
        
        return trace
    
    module.HTTPError.set_response = wrapper(module.HTTPError.set_response)
